__all__ = ()

from scarletio import include

from ...emoji import Emoji

from .constants import (
    COMPONENT_CUSTOM_ID_LENGTH_MAX, COMPONENT_LABEL_LENGTH_MAX, COMPONENT_MAX_LENGTH_MAX, COMPONENT_MAX_LENGTH_MIN,
    COMPONENT_MIN_LENGTH_MAX, COMPONENT_MIN_LENGTH_MIN, COMPONENT_OPTION_LENGTH_MAX, COMPONENT_OPTION_LENGTH_MIN,
    COMPONENT_OPTION_MAX_VALUES_MAX, COMPONENT_OPTION_MAX_VALUES_MIN, COMPONENT_OPTION_MIN_VALUES_MAX,
    COMPONENT_OPTION_MIN_VALUES_MIN, COMPONENT_SUB_COMPONENT_LIMIT, COMPONENT_TITLE_LENGTH_MAX,
    COMPONENT_TITLE_LENGTH_MIN
)
from .preinstanced import ComponentType


ComponentSelectOption = include('ComponentSelectOption')
ComponentBase = include('ComponentBase')

COMPONENT_TYPE_ROW = ComponentType.row


def _debug_component_components(components):
    """
    Checks whether given `component.components` value is correct.
    
    Parameters
    ----------
    components : `None` or (`list`, `tuple`) of ``ComponentBase``
        Sub-components.
    
    Raises
    ------
    AssertionError
        - If `components`'s length is out of the expected range [0:5].
        - If `components` is neither `None`, `tuple`, `list`.
        - If `components` contains a non ``ComponentBase``.
    """
    if (components is None):
        pass
    elif isinstance(components, (tuple, list)):
        if (len(components) > COMPONENT_SUB_COMPONENT_LIMIT):
            raise AssertionError(
                f'`component.components` can have maximum 5 sub-components, got '
                f'{len(components)}; {components!r}.'
            )
        
        for component in components:
            if not isinstance(component, ComponentBase):
                raise AssertionError(
                    f'`component` can be `{ComponentBase.__name__}`, got '
                    f'{component.__class__.__name__}.'
                )
            
            if component.type is COMPONENT_TYPE_ROW:
                raise AssertionError(
                    f'Cannot add `{COMPONENT_TYPE_ROW}` type as sub components, got '
                    f'{component!r}.'
                )
    else:
        raise AssertionError(
            f'`components` can be `None`, `tuple`, `list`, got '
            f'{components.__class__.__name__}; {components!r}.'
        )


def _debug_component_custom_id(custom_id):
    """
    Checks whether given `component.custom_id` value is correct.
    
    Parameters
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which button was clicked by the user.
    
    Raises
    ------
    AssertionError
        - If `custom_id` was not given neither as `None`, `str`.
        - If `custom_id`'s length is over `100`.
    """
    if (custom_id is None):
        pass
    elif isinstance(custom_id, str):
        custom_id_length = len(custom_id)
        if custom_id_length == 0:
            raise AssertionError(
                f'`custom_id` is not nullable.'
            )
        
        if custom_id_length > COMPONENT_CUSTOM_ID_LENGTH_MAX:
            raise AssertionError(
                f'`custom_id`\'s max length is {COMPONENT_CUSTOM_ID_LENGTH_MAX!r}, got '
                f'{len(custom_id)!r}; {custom_id!r}.'
            )
    else:
        raise AssertionError(
            f'`custom_id` can be `None`, `str`, got {custom_id.__class__.__name__}; {custom_id!r}.'
        )


def _debug_component_emoji(emoji):
    """
    Checks whether the given `component.emoji` value is correct.
    
    Parameters
    ----------
    emoji : `None`, ``Emoji``
        Emoji of the button if applicable.
    
    Raises
    ------
    AssertionError
        If `emoji` was not given as ``Emoji``.
    """
    if emoji is None:
        pass
    elif isinstance(emoji, Emoji):
        pass
    else:
        raise AssertionError(f'`emoji` can be `{Emoji.__name__}`, got '
            f'{emoji.__class__.__name__}')


def _debug_component_label(label):
    """
    Checks whether the given `component.label` value is correct.
    
    Parameters
    ----------
    label : `None`, `str`
        Label of the component.
    
    Raises
    ------
    AssertionError
        - If `label` was not given neither as `None` nor as `int`.
        - If `label`'s length is over `80`.
    """
    if label is None:
        pass
    elif isinstance(label, str):
        if len(label) > COMPONENT_LABEL_LENGTH_MAX:
            raise AssertionError(
                f'`label`\'s max length can be {COMPONENT_LABEL_LENGTH_MAX!r}, got '
                f'{len(label)!r}; {label!r}.'
            )
    else:
        raise AssertionError(
            f'`label` can be `None`, `str`, got {label.__class__.__name__}; {label!r}.'
        )

    
def _debug_component_title(title):
    """
    Checks whether the given `component.title` value is correct.
    
    Parameters
    ----------
    title : `None`, `str`
        Title of the component.
    
    Raises
    ------
    AssertionError
        - If `title` was not given neither as `None` nor as `int`.
        - If `title`'s length is out of the expected range.
    """
    if title is None:
        raise AssertionError(
            f'`title`\'s is not nullable.'
        )
        
    elif isinstance(title, str):
        title_length = len(title)
        
        if title_length < COMPONENT_TITLE_LENGTH_MIN:
            raise AssertionError(
                f'`title`\'s minimal length is {COMPONENT_TITLE_LENGTH_MIN!r}, got {title_length!r}; {title!r}.'
            )
        
        if title_length > COMPONENT_TITLE_LENGTH_MAX:
            raise AssertionError(
                f'`title`\'s maximal length is {COMPONENT_TITLE_LENGTH_MAX!r}, got {title_length!r}; {title!r}.'
            )
    
    else:
        raise AssertionError(
            f'`title` can be `None`, `str`, got {title.__class__.__name__}; {title!r}.'
        )


def _debug_component_enabled(enabled):
    """
    Checks whether the given `component.enabled` value is correct.
    
    Parameters
    ----------
    enabled : `bool`
        Whether the button is enabled.
    
    Raises
    ------
    AssertionError
        If `enabled` was not given as `bool`.
    """
    if not isinstance(enabled, bool):
        raise AssertionError(
            f'`enabled` can be `bool`, got {enabled.__class__.__name__}; {enabled!r}.'
        )


def _debug_component_url(url):
    """
    Checks whether the given `component.url` value is correct.
    
    Parameters
    ----------
    url : `None`, `str`
        Url to redirect to when clicking on a button.
    
    Raises
    ------
    AssertionError
        If `url` was not given neither as `None`, `str`.
    """
    if url is None:
        pass
    elif isinstance(url, str):
        pass
    else:
        raise AssertionError(
            f'`url` can be `None`, `str`, got {url.__class__.__name__}; {url!r}.'
        )


def _debug_component_select_option_value(value):
    """
    Checks whether the given `component_option.value` value is correct.
    
    Parameters
    ----------
    value : `str`
        A component option's value.
    
    Raises
    ------
    AssertionError
        If `value` was not given as `str`.
    """
    if not isinstance(value, str):
        raise AssertionError(
            f'`value` can be `str`, got {value.__class__.__name__}; {value!r}.'
        )


def _debug_component_initial_value(value):
    """
    Checks whether the given `component_option.value` value is correct.
    
    Parameters
    ----------
    value : `str`
        A component option's value.
    
    Raises
    ------
    AssertionError
        If `value` was not given as `str`.
    """
    if not isinstance(value, str):
        raise AssertionError(
            f'`value` can be `str`, got {value.__class__.__name__}; {value!r}.'
        )


def _debug_component_description(description):
    """
    Checks whether the given `component_option.description` value is correct.
    
    Parameters
    ----------
    description : `None`, `str`
        A component option's description.
    
    Raises
    ------
    AssertionError
        If `description` was not given neither as `None`, `str`.
    """
    if description is None:
        pass
    elif isinstance(description, str):
        pass
    else:
        raise AssertionError(
            f'`description` can be `None`, `str`, got {description.__class__.__name__}; {description!r}.'
        )


def _debug_component_default(default):
    """
    Checks whether the given `component_option.default` value is correct.
    
    Parameters
    ----------
    default : `bool`
        Whether this component option is the default one.
    
    Raises
    ------
    AssertionError
        If `default` was not given as `bool`.
    """
    if not isinstance(default, bool):
        raise AssertionError(
            f'`default` can be `bool`, got {default.__class__.__name__}; {default!r}.'
        )


def _debug_component_options(options):
    """
    Checks whether given `component.options` value is correct.
    
    Parameters
    ----------
    options : `None` or (`list`, `tuple`) of ``ComponentSelectOption``
        Sub-options.
    
    Raises
    ------
    AssertionError
        - If `options` is neither `None`, `tuple`, `list`.
        - If `options` contains a non ``ComponentSelectOption``.
        - If `options`'s length is out of the expected [1:25] range.
    """
    if options is None:
        option_length = 0
    elif isinstance(options, (tuple, list)):
        for option in options:
            if not isinstance(option, ComponentSelectOption):
                raise AssertionError(
                    f'`option` can be `{ComponentSelectOption.__name__}`, got '
                    f'{option.__class__.__name__}, {option!r}.'
                )
        
        option_length = len(options)
    else:
        raise AssertionError(
            f'`options` can be `None`, `tuple`, `list`, got {options.__class__.__name__}; {options!r}.'
        )
    
    if (option_length < COMPONENT_OPTION_LENGTH_MIN) or (option_length > COMPONENT_OPTION_LENGTH_MAX):
        raise AssertionError(
            f'`options`\'s length can be in range '
            f'[{COMPONENT_OPTION_LENGTH_MIN}:{COMPONENT_OPTION_LENGTH_MAX}], got {option_length}; {options!r}.'
        )


def _debug_component_placeholder(placeholder):
    """
    Checks whether the given `component_option.placeholder` value is correct.
    
    Parameters
    ----------
    placeholder : `None`, `str`
        The placeholder text of a component select.
    
    Raises
    ------
    AssertionError
        - If `placeholder` is neither `None` nor `str`.
    """
    if placeholder is None:
        pass
    elif isinstance(placeholder, str):
        pass
    else:
        raise AssertionError(
            f'`placeholder` can be `None`, `str`, got {placeholder.__class__.__name__}; {placeholder!r}.'
        )


def _debug_component_min_values(min_values):
    """
    Checks whether the given `component_option.min_values` value is correct.
    
    Parameters
    ----------
    min_values : `int`
        The min values of a component select.
    
    Raises
    ------
    AssertionError
        - If `min_values` was not given as `int`.
        - If `min_values`'s is out of range [0:15].
    """
    if not isinstance(min_values, int):
        raise AssertionError(
            f'`min_values` can be `int`, got {min_values.__class__.__name__}; {min_values!r}.'
        )
    
    if (min_values < COMPONENT_OPTION_MIN_VALUES_MIN) or (min_values > COMPONENT_OPTION_MIN_VALUES_MAX):
        raise AssertionError(
            f'`min_values` can be in range '
            f'[{COMPONENT_OPTION_MIN_VALUES_MIN}:{COMPONENT_OPTION_MAX_VALUES_MIN}], got {min_values!r}.'
        )


def _debug_component_max_values(max_values):
    """
    Checks whether the given `component_option.max_values` value is correct.
    
    Parameters
    ----------
    max_values : `int`
        The max values of a component select.
    
    Raises
    ------
    AssertionError
        - If `max_values` was not given as `int`.
        - If `max_values`'s is out of range [1:25].
    """
    if not isinstance(max_values, int):
        raise AssertionError(
            f'`max_values` can be `int`, got {max_values.__class__.__name__}; {max_values!r}.'
        )

    if (max_values < COMPONENT_OPTION_MAX_VALUES_MIN) or (max_values > COMPONENT_OPTION_MAX_VALUES_MAX):
        raise AssertionError(
            f'`max_values` can be in range '
            f'[{COMPONENT_OPTION_MAX_VALUES_MIN}:{COMPONENT_OPTION_MAX_VALUES_MAX}], got {max_values!r}.'
        )


def _debug_component_max_length(max_length):
    """
    Checks whether the given `component_text_input.max_length` value is correct.
    
    Parameters
    ----------
    max_length : `int`
        The max values of a component select.
    
    Raises
    ------
    AssertionError
        - If `max_length` was not given as `int`.
        - If `man_length`'s is out of the expected range.
    """
    if not isinstance(max_length, int):
        raise AssertionError(
            f'`max_length` can be `int`, got {max_length.__class__.__name__}; {max_length!r}.'
        )

    if (max_length < COMPONENT_MAX_LENGTH_MIN) or (max_length > COMPONENT_MAX_LENGTH_MAX):
        raise AssertionError(
            f'`max_length` can be in range '
            f'[{COMPONENT_MAX_LENGTH_MIN}:{COMPONENT_MAX_LENGTH_MAX}], got {max_length!r}.'
        )


def _debug_component_min_length(min_length):
    """
    Checks whether the given `component_text_input.min_length` value is correct.
    
    Parameters
    ----------
    min_length : `int`
        The max values of a component select.
    
    Raises
    ------
    AssertionError
        - If `min_length` was not given as `int`.
        - If `min_length`'s is out of the expected range.
    """
    if not isinstance(min_length, int):
        raise AssertionError(
            f'`min_length` can be `int`, got {min_length.__class__.__name__}; {min_length!r}.'
        )

    if (min_length < COMPONENT_MIN_LENGTH_MIN) or (min_length > COMPONENT_MIN_LENGTH_MAX):
        raise AssertionError(
            f'`min_length` can be in range '
            f'[{COMPONENT_MIN_LENGTH_MIN}:{COMPONENT_MIN_LENGTH_MAX}], got {min_length!r}.'
        )


def _debug_component_required(required):
    """
    Checks whether the given `component_text_input.required` value is correct.
    
    Parameters
    ----------
    required : `bool`
        Whether this component text input is required.
    
    Raises
    ------
    AssertionError
        If `required` was not given neither as `None`, `bool`.
    """
    if (required is not None) and (not isinstance(required, bool)):
        raise AssertionError(
            f'`required` can be `None`, `bool`, got {required.__class__.__name__}; {required!r}.'
        )


def _debug_component_text_input_value(value):
    """
    Checks whether the given `component_text_input.value` value is correct.
    
    Parameters
    ----------
    value : `None`, `str`
        The default value of the text input.
    
    Raises
    ------
    AssertionError
        If `value` was not given neither as `None`, `str`.
    """
    if (value is not None) and (not isinstance(value, str)):
        raise AssertionError(
            f'`value` can be `None`, `str`, got {value.__class__.__name__}; {value!r}.'
        )




__all__ = ('ApplicationCommandInteraction', 'ApplicationCommandInteractionOption')

from scarletio import copy_docs

from ...bases import DiscordEntity
from ...channel import create_partial_channel_from_data
from ...message import Attachment, Message
from ...role import Role
from ...user import User

from ..application_command.preinstanced import ApplicationCommandOptionType

from .interaction_field_base import InteractionFieldBase


class ApplicationCommandInteraction(DiscordEntity, InteractionFieldBase):
    """
    Represents an ``ApplicationCommand`` invoked by a user.
    
    Attributes
    ----------
    id : `int`
        The represented application command's identifier number.
    name : `str`
        The name of the command. It's length can be in range [1:32].
    options : `None`, `tuple` of ``ApplicationCommandInteractionOption``
        The parameters and values from the user if any. Defaults to `None` if non is received.
    resolved_attachments : `None`, `dict` of (`int`, `Attachment``) items
        Resolved received attachments stored by their identifier as keys if any.
    resolved_channels : `None`, `dict` of (`int`, ``Channel``) items
        Resolved received channels stored by their identifier as keys if any.
    resolved_roles : `None`, `dict` of (`int`, ``Role``) items
        Resolved received roles stored by their identifier as keys if any.
    resolved_messages : `None`, `dict` of (`int`, ``Message``) items
        Resolved received messages stored by their identifier as keys if any.
    resolved_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        Resolved received users stored by their identifier as keys if any.
    target_id : `int`
        The interaction's target's identifier.
    """
    __slots__ = (
        'name', 'options', 'resolved_attachments', 'resolved_channels', 'resolved_roles', 'resolved_messages',
        'resolved_users', 'target_id'
    )
    
    
    @copy_docs(InteractionFieldBase.__eq__)
    def __new__(cls, data, interaction_event):
        # id
        id_ = int(data['id'])
        
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(ApplicationCommandInteractionOption(option_data) for option_data in option_datas)
        
        # resolved_channels & resolved_roles & resolved_messages & resolved_users
        try:
            resolved_data = data['resolved']
        except KeyError:
            resolved_attachments = None
            resolved_channels = None
            resolved_roles = None
            resolved_messages = None
            resolved_users = None
        else:
            # resolved_attachments
            try:
                resolved_attachment_datas = resolved_data['attachments']
            except KeyError:
                resolved_attachments = None
            else:
                if resolved_attachment_datas:
                    resolved_attachments = {}
                    
                    for attachment_data in resolved_attachment_datas.values():
                        attachment = Attachment(attachment_data)
                        if (attachment is not None):
                            resolved_attachments[attachment.id] = attachment
                    
                    if not resolved_attachments:
                        resolved_attachments = None
                else:
                    resolved_attachments = None
            
            # resolved_channels
            try:
                resolved_channel_datas = resolved_data['channels']
            except KeyError:
                resolved_channels = None
            else:
                if resolved_channel_datas:
                    resolved_channels = {}
                    
                    for channel_data in resolved_channel_datas.values():
                        channel = create_partial_channel_from_data(channel_data, interaction_event.id)
                        if (channel is not None):
                            resolved_channels[channel.id] = channel
                    
                    if not resolved_channels:
                        resolved_channels = None
                else:
                    resolved_channels = None
            
            # resolved_roles
            try:
                resolved_role_datas = resolved_data['roles']
            except KeyError:
                resolved_roles = None
            else:
                if resolved_role_datas:
                    resolved_roles = {}
                    for role_data in resolved_role_datas.values():
                        role = Role(role_data, interaction_event.guild_id)
                        resolved_roles[role.id] = role
                else:
                    resolved_roles = None
            
            # resolved_messages
            try:
                resolved_message_datas = resolved_data['messages']
            except KeyError:
                resolved_messages = None
            else:
                if resolved_message_datas:
                    resolved_messages = {}
                    
                    for message_data in resolved_message_datas.values():
                        message = Message(message_data)
                        resolved_messages[message.id] = message
                else:
                    resolved_messages = None
            
            # resolved_users
            try:
                resolved_user_datas = resolved_data['users']
            except KeyError:
                resolved_users = None
            else:
                if resolved_user_datas:
                    try:
                        resolved_guild_profile_datas = resolved_data['members']
                    except KeyError:
                        resolved_guild_profile_datas = None
                    
                    resolved_users = {}
                    
                    for user_id, user_data in resolved_user_datas.items():
                        if resolved_guild_profile_datas is None:
                            guild_profile_data = None
                        else:
                            guild_profile_data = resolved_guild_profile_datas.get(user_id, None)
                        
                        user = User.from_data(user_data, guild_profile_data, interaction_event.guild_id)
                        resolved_users[user.id] = user
                        
                        if (guild_profile_data is not None):
                            interaction_event._add_cached_user(user)
                
                else:
                    resolved_users = None
        
        # target_id
        target_id = data.get('target_id', None)
        if target_id is None:
            target_id = 0
        else:
            target_id = int(target_id)
        
        self = object.__new__(cls)
        self.id = id_
        self.name = name
        self.options = options
        self.resolved_attachments = resolved_attachments
        self.resolved_users = resolved_users
        self.resolved_channels = resolved_channels
        self.resolved_roles = resolved_roles
        self.resolved_messages = resolved_messages
        self.target_id = target_id
        
        interaction_event._add_response_waiter()
        
        return self
    
    
    @copy_docs(InteractionFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', name=', repr(self.name),
        ]
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        target = self.target
        if (target is not None):
            repr_parts.append(', target=')
            repr_parts.append(repr(target))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(InteractionFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # resolved_attachments
        if self.resolved_attachments != other.resolved_attachments:
            return False
        
        # resolved_channels
        if self.resolved_channels != other.resolved_channels:
            return False
        
        # resolved_roles
        if self.resolved_roles != other.resolved_roles:
            return False
        
        # resolved_messages
        if self.resolved_messages != other.resolved_messages:
            return False
        
        # resolved_users
        if self.resolved_users != other.resolved_users:
            return False
        
        # target_id
        if self.target_id != other.target_id:
            return False
        
        return True
    
    
    @copy_docs(InteractionFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options)
            
            for option in options:
                hash_value ^= hash(option)
        
        # resolved_attachments
        resolved_attachments = self.resolved_attachments
        if (resolved_attachments is not None):
            hash_value ^= (len(resolved_attachments) << 24)
            
            for attachment_id in resolved_attachments.keys():
                hash_value ^= attachment_id
        
        # resolved_channels
        resolved_channels = self.resolved_channels
        if (resolved_channels is not None):
            hash_value ^= (len(resolved_channels) << 8)
            
            for channel_id in resolved_channels.keys():
                hash_value ^= channel_id
        
        # resolved_roles
        resolved_roles = self.resolved_roles
        if (resolved_roles is not None):
            hash_value ^= (len(resolved_roles) << 12)
            
            for role_id in resolved_roles.keys():
                hash_value ^= role_id
        
        # resolved_messages
        resolved_messages = self.resolved_messages
        if (resolved_messages is not None):
            hash_value ^= (len(resolved_messages) << 16)
            
            for message_id in resolved_messages.keys():
                hash_value ^= message_id
        
        # resolved_users
        resolved_users = self.resolved_users
        if (resolved_users is not None):
            hash_value ^= (len(resolved_users) << 20)
            
            for user_id in resolved_users.keys():
                hash_value ^= user_id
        
        # target_id
        hash_value ^= self.target_id
        
        return hash_value
    
    
    def resolve_entity(self, entity_id):
        """
        Tries to resolve the entity by the given identifier.
        
        Parameters
        ----------
        entity_id : ``int``
            The entity's identifier.
        
        Returns
        -------
        resolved : `None`, ``DiscordEntity``
            The resolved discord entity if found.
        """
        # Is used at `InteractionEvent.target`, which wanna access user and message first, so we check that two first.
        resolved_messages = self.resolved_messages
        if (resolved_messages is not None):
            try:
                entity = resolved_messages[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_users = self.resolved_users
        if (resolved_users is not None):
            try:
                entity = resolved_users[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_roles = self.resolved_roles
        if (resolved_roles is not None):
            try:
                entity = resolved_roles[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_channels = self.resolved_channels
        if (resolved_channels is not None):
            try:
                entity = resolved_channels[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        resolved_attachments = self.resolved_attachments
        if (resolved_attachments is not None):
            try:
                entity = resolved_attachments[entity_id]
            except KeyError:
                pass
            else:
                return entity
        
        
        return None
    
    @property
    def target(self):
        """
        Returns the interaction event's target.
        
        Only applicable for context application commands.
        
        Returns
        -------
        target : ``ClientUserBase``, ``Message``
        """
        target_id = self.target_id
        if target_id:
            return self.resolve_entity(target_id)


class ApplicationCommandInteractionOption:
    """
    Represents an option of a ``ApplicationCommandInteraction``.
    
    Attributes
    ----------
    name : `str`
        The option's name.
    
    options : `None`, `tuple` of ``ApplicationCommandInteractionOption``
        The parameters and values from the user. Present if a sub-command was used. Defaults to `None` if non is
        received.
        
        Mutually exclusive with the `value` field.
    
    type : ``ApplicationCommandOptionType``
        The option's type.
    
    value : `None`, `str`, `bool`, `float`, `int`
        The given value by the user.
        
        Mutually exclusive with the `options` field,
    
    """
    __slots__ = ('name', 'options', 'type', 'value')
    
    def __new__(cls, data):
        """
        Creates a new ``ApplicationCommandInteractionOption`` from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command interaction option data.
        """
        # name
        name = data['name']
        
        # options
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = tuple(ApplicationCommandInteractionOption(option_data) for option_data in option_datas)
        
        # type
        type_ = ApplicationCommandOptionType.get(data.get('type', 0))
        
        # value
        value = data.get('value', None)
        
        self = object.__new__(cls)
        
        self.name = name
        self.options = options
        self.type = type_
        self.value = value
        
        return self
    
    
    def __repr__(self):
        """Returns the application command interaction option's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ', name=', repr(self.name),
        ]
        
        type_ = self.type
        if type_ is not ApplicationCommandOptionType.none:
            repr_parts.append(', type=')
            repr_parts.append(type_.name)
            repr_parts.append(' (')
            repr_parts.append(repr(type_.value))
            repr_parts.append(')')
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two application command interaction options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # name
        if self.name != other.name:
            return False
        
        # options
        if self.options != other.options:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command interaction option's hash value."""
        hash_value = 0
        
        # name
        hash_value ^= hash(self.name)
        
        # options
        options = self.options
        if (options is not None):
            hash_value ^= len(options)
            
            for option in options:
                hash_value ^= hash(option)
        
        # type
        hash_value ^= (self.type.value << 8)
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value



from .component_base import *
from .component_button import *
from .component_dynamic import *
from .component_row import *
from .component_select import *
from .component_select_option import *
from .component_text_input import *
from .constants import *
from .debug import *
from .form import *
from .preinstanced import *
from .utils import *

__all__ = (
    *component_base.__all__,
    *component_button.__all__,
    *component_dynamic.__all__,
    *component_row.__all__,
    *component_select.__all__,
    *component_select_option.__all__,
    *component_text_input.__all__,
    *constants.__all__,
    *debug.__all__,
    *form.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)



__all__ = ()

import sys
from importlib.util import spec_from_file_location
from os.path import (
    exists, isdir as is_directory, isfile as is_file, join as join_paths, sep as PATH_SEPARATOR, split as split_path
)


def create_spec_if_exists(full_name, path):
    """
    Creates module spec if the path exists.
    
    Parameters
    ----------
    full_name : `str`
        The full name of the module to find.
    path : `str`
        A path to check whether it represents a python file / directory with `__init__.py`.
    
    Returns
    -------
    spec : `None`, `ModuleSpec`
    """
    if is_directory(path):
        path = join_paths(path, '__init__.py')
        if exists(path):
            return spec_from_file_location(full_name, path)
        
        return None
    
    path += '.py'
    
    if exists(path) and is_file(path):
        return spec_from_file_location(full_name, path)
    
    return None


def try_find_spec_in_system_paths(full_name, full_name_parts, path):
    """
    Tries to the find the spec in the given path. uses `sys.path` to default to when doing so.
    
    Parameters
    ----------
    full_name : `str`
        The full name of the module to find.
    full_name_parts : `list` of `str`
        `full_name` separated by `'.'`.
    path : `str`
        Path to find the module in.
    
    Returns
    -------
    spec : `None`, `ModuleSpec`
    """
    for system_path in sys.path:
        if system_path.endswith(PATH_SEPARATOR):
            compare_system_path = system_path
        else:
            compare_system_path = system_path + PATH_SEPARATOR
        
        if not path.startswith(compare_system_path):
            continue
        
        spec = create_spec_if_exists(full_name, join_paths(system_path, *full_name_parts))
        if (spec is not None):
            return spec
    
    return None


def try_find_spec_in_path_fallback(full_name, full_name_parts, path):
    """
    Tries to the find the spec in the given path.
    
    This is a fallback function and only called if every other way fails.
    
    Parameters
    ----------
    full_name : `str`
        The full name of the module to find.
    full_name_parts : `list` of `str`
        `full_name` separated by `'.'`.
    path : `str`
        Path to find the module in.
    
    Returns
    -------
    spec : `None`, `ModuleSpec`
    """
    first_name_part = full_name_parts[0]
    
    while True:
        parent_name, directory_name = split_path(path)
        if not directory_name:
            return None
        
        if directory_name == first_name_part:
            module_path = join_paths(parent_name, *full_name_parts)
            spec = create_spec_if_exists(full_name, module_path)
            if (spec is not None):
                return spec
        
        path = parent_name
        continue


def find_spec_in_paths(full_name, paths):
    """
    Tries to the find the spec in the given paths.
    
    Parameters
    ----------
    full_name : `str`
        The full name of the module to find.
    paths : `list` of `str`
        Path to find the module in.
    
    Returns
    -------
    spec : `None`, `ModuleSpec`
    """
    full_name_parts = full_name.split('.')
    if not full_name_parts:
        return None
    
    if paths is None:
        for system_path in sys.path:
            module_path = join_paths(system_path, *full_name_parts)
            spec = create_spec_if_exists(full_name, module_path)
            if (spec is not None):
                return spec
        
    else:
        for path in paths:
            spec = try_find_spec_in_system_paths(full_name, full_name_parts, path)
            if (spec is not None):
                return spec
        
        for path in paths:
            spec = try_find_spec_in_path_fallback(full_name, full_name_parts, path)
            if (spec is not None):
                return spec
    
    return None



__all__ = ('Application', )

import warnings

from ..bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ..core import APPLICATIONS
from ..http import urls as module_urls
from ..preconverters import preconvert_bool, preconvert_flag, preconvert_snowflake, preconvert_str
from ..user import ClientUserBase, User, ZEROUSER
from .embedded_activity_configuration import EmbeddedActivityConfiguration
from .flags import ApplicationFlag
from .miscellaneous import ApplicationExecutable, ApplicationInstallParameters, ApplicationSubEntity, ThirdPartySKU
from .team import Team


class Application(DiscordEntity, immortal=True):
    """
    Represents a Discord application with all of it's spice.
    
    When a ``Client`` is created, it starts it's life with an empty application by default. However when the client
    logs in, it's application is requested, but it can be updated by ``Client.update_application_info`` anytime.
    
    Attributes
    ----------
    aliases : `None`, `list` of `str`
        Aliases of the application's name. Defaults to `None`.
    bot_public : `bool`.
        Whether not only the application's owner can join the application's bot to guilds. Defaults to `False`
    bot_require_code_grant : `bool`
        Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        Defaults to `False`.
    cover_hash : `int`
        The application's store cover image's hash in `uint128`. If the application is sold at Discord, this image
        will be used at the store.
    cover_type : ``IconType``
        The application's store cover image's type.
    custom_install_url : `None`, `str`
        The application's default custom authorization link, if enabled.
    description : `str`
        The description of the application. Defaults to empty string.
    developers : `None`, `list` of ``ApplicationSubEntity``
        A list of the application's games' developers. Defaults to `None`.
    embedded_activity_configuration : `None`, EmbeddedActivityConfiguration
        Configuration for the embedded activity.
    eula_id : `int`
        The end-user license agreement's id of the application. Defaults to `0` if not applicable.
    executables : `None`, `list` of ``ApplicationExecutable``
        A list of the application's executables. Defaults to `None`.
    flags : ``ApplicationFlag``
        The application's public flags.
    guild_id : `int`
        If the application is a game sold on Discord, this field tells in which guild it is.
        Defaults to `0` if not applicable.
    hook : `bool`
        Defaults to `False`.
    install_parameters : `None`, ``ApplicationInstallParameters``
        Settings for the application's default in-app authorization link, if enabled.
    icon_hash : `int`
        The application's icon's hash as `uint128`.
    icon_type : ``IconType``
        The application's icon's type.
    id : `int`
        The application's id. Defaults to `0`. Meanwhile set as `0`, hashing the application will raise `RuntimeError`.
    max_participants : `int`
        The maximal amount of users, who can join the application's embedded activity. Defaults to `0`.
    name : `str`
        The name of the application. Defaults to empty string.
    overlay : `bool`
        Defaults to `False`.
    overlay_compatibility_hook : `bool`
        Defaults to `False`.
    owner : ``ClientUserBase``, ``Team``
        The application's owner. Defaults to `ZEROUSER`.
    primary_sku_id : `int`
        If the application is a game sold on Discord, this field will be the id of the created `Game SKU`.
        Defaults to `0`.
    privacy_policy_url : `None`, `str`
        The url of the application's privacy policy. Defaults to `None`.
    publishers : `None`, `list` of ``ApplicationSubEntity``
        A list of the application's games' publishers. Defaults to `None`.
    rpc_origins : `None`, `list` of `str`
        A list of `rpc` origin urls, if `rpc` is enabled. Set as `None` if would be an empty list.
    slug : `None`, `str`
        If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        Defaults to `None`.
    splash_hash : `int`
        The application's splash image's hash as `uint128`.
    splash_type : ``IconType``
        The application's splash image's type.
    tags : `None`, `tuple` of `str`
        Up to 5 tags describing the content and functionality of the application.
    terms_of_service_url : `None`, `str`
        The url of the application's terms of service. Defaults to `None`.
    third_party_skus : `None`, `list` of ``ThirdPartySKU``
         A list of the application's third party stock keeping units. Defaults to `None`.
    verify_key : `str`
        A base64 encoded key for the GameSDK's `GetTicket`. Defaults to empty string.
    
    Notes
    -----
    The instances of the class support weakreferencing.
    """
    __slots__ = (
        'aliases', 'bot_public', 'bot_require_code_grant', 'custom_install_url', 'description', 'developers',
        'embedded_activity_configuration', 'eula_id', 'executables', 'flags', 'guild_id', 'hook', 'install_parameters',
        'max_participants', 'name', 'overlay', 'overlay_compatibility_hook', 'owner', 'primary_sku_id',
        'privacy_policy_url', 'publishers', 'rpc_origins', 'slug', 'tags', 'terms_of_service_url', 'third_party_skus',
        'verify_key'
    )
    
    cover = IconSlot(
        'cover',
        'cover_image',
        module_urls.application_cover_url,
        module_urls.application_cover_url_as,
        add_updater = False,
    )
    icon = IconSlot(
        'icon',
        'icon',
        module_urls.application_icon_url,
        module_urls.application_icon_url_as,
        add_updater = False,
    )
    splash = IconSlot(
        'splash',
        'splash',
        None,
        None,
        add_updater = False,
    )
    
    @classmethod
    def _create_empty(cls, application_id):
        """
        Creates an empty application, with it's default attributes set.
        
        Parameters
        ----------
        application_id : `int`
            The application's default identifier.
        
        Returns
        -------
        self : ``Application``
            The created application.
        """
        self = object.__new__(cls)
        
        self.aliases = None
        self.bot_public = False
        self.bot_require_code_grant = False
        self.cover_hash = 0
        self.cover_type = ICON_TYPE_NONE
        self.custom_install_url = None
        self.description = ''
        self.developers = None
        self.embedded_activity_configuration = None
        self.eula_id = 0
        self.executables = None
        self.flags = ApplicationFlag()
        self.guild_id = None
        self.hook = None
        self.install_parameters = None
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        self.id = application_id
        self.max_participants = 0
        self.name = ''
        self.overlay = False
        self.overlay_compatibility_hook = False
        self.owner = ZEROUSER
        self.primary_sku_id = 0
        self.privacy_policy_url = None
        self.publishers = None
        self.rpc_origins = None
        self.slug = None
        self.splash_hash = 0
        self.splash_type = ICON_TYPE_NONE
        self.terms_of_service_url = None
        self.tags = None
        self.third_party_skus = None
        self.verify_key = ''
        
        return self
    
    
    def _create_update(self, data, ready_data):
        """
        Creates a new application if the given data refers to an other one. Updates the application and returns it.
        
        Parameters
        ----------
        data : or `dict` of (`str`, `Any`) items, Optional
            Application data received from Discord.
        ready_data : `bool`
            Whether the application data was received from a ready event.
        
        Returns
        -------
        self : ``Application``
            The created or updated application.
        """
        application_id = int(data['id'])
        
        if self.id == 0:
            try:
                self = APPLICATIONS[application_id]
            except KeyError:
                self.id = application_id
                APPLICATIONS[application_id] = self
        
        if ready_data:
            try:
                flags = data['flags']
            except KeyError:
                flags = ApplicationFlag()
            else:
                flags = ApplicationFlag(flags)
            self.flags = flags
        else:
            self._update_attributes(data, set_owner=True)
        
        return self
    
    def __new__(cls, data):
        """
        Creates a new application with the given data. If the application already exists, updates it and returns that
        instead.
        
        Parameters
        ----------
        data : or `dict` of (`str`, `Any`) items, Optional
            Application data received from Discord.
        
        Returns
        -------
        self : `˙Application``
        """
        application_id = int(data['id'])
        try:
            self = APPLICATIONS[application_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = application_id
            self.flags = ApplicationFlag()
            self._update_attributes(data, set_owner=True)
        else:
            self._update_attributes(data)
        
        return self
    
    def __hash__(self):
        """Returns the application's hash value."""
        id_ = self.id
        if id_:
            return id_
        
        raise TypeError(f'Cannot hash partial {self.__class__.__name__} object.')
    
    @property
    def partial(self):
        """
        Returns whether the application is partial.
        
        An application if partial, if it's id is set as `0`.
        Returns
        -------
        partial : `bool`
        """
        return (self.id == 0)
    
    def __repr__(self):
        """Returns the application's representation"""
        result = [
            '<',
            self.__class__.__name__,
        ]
        
        id_ = self.id
        if id_:
            result.append(' id=')
            result.append(repr(id_))
            result.append(', name=')
            result.append(repr(self.name))
        else:
            result.append(' partial')
        
        result.append('>')
        
        return ''.join(result)
    
    
    def _update_attributes(self, data, set_owner=False):
        """
        Updates the application with the data received from Discord.
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `Any`) items
            Application data received from Discord.
        set_owner : `bool` = `False`, Optional
            Whether the application's owner should be set from the given data.
            Should be given as `True`, if the application is created or if it contains owner data.
        """
        self.name = data['name']
        
        self.description = data['description']
        
        rpc_origins = data.get('rpc_origins', None)
        if (rpc_origins is not None) and (not rpc_origins):
            rpc_origins = None
        
        self.rpc_origins = rpc_origins
        
        self.bot_public = data.get('bot_public', False)
        self.bot_require_code_grant = data.get('bot_require_code_grant', False)
        self.verify_key = data['verify_key']
        
        if set_owner:
            team_data = data.get('team', None)
            if team_data is None:
                owner_data = data.get('owner', None)
                if owner_data is None:
                    owner = ZEROUSER
                else:
                    owner = User.from_data(owner_data)
            else:
                owner = Team(team_data)
            
            self.owner = owner
        
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self.guild_id = guild_id
        
        primary_sku_id = data.get('primary_sku_id', None)
        if primary_sku_id is None:
            primary_sku_id = None
        else:
            primary_sku_id = int(primary_sku_id)
        self.primary_sku_id = primary_sku_id
        
        self.slug = data.get('slug', None)
        
        self._set_cover(data)
        self._set_icon(data)
        
        developers_data = data.get('developers', None)
        if (developers_data is None) or (not developers_data):
            developers = None
        else:
            developers = [ApplicationSubEntity(developer_data) for developer_data in developers_data]
        
        self.developers = developers
        
        self.hook = data.get('hook', False)
        
        publishers_data = data.get('publishers', None)
        if (publishers_data is None) or (not publishers_data):
            publishers = None
        else:
            publishers = [ApplicationSubEntity(publisher_data) for publisher_data in publishers_data]
        
        self.publishers = publishers
        
        executables_data = data.get('executables', None)
        if (executables_data is None) or (not executables_data):
            executables = None
        else:
            executables = [ApplicationExecutable(executable_data) for executable_data in executables_data]
        
        self.executables = executables
        
        self._set_splash(data)
        
        third_party_skus_data = data.get('third_party_skus', None)
        if (third_party_skus_data is None) or (not third_party_skus_data):
            third_party_skus = None
        else:
            third_party_skus = [ThirdPartySKU(third_party_sku_data) for third_party_sku_data in third_party_skus_data]
        
        self.third_party_skus = third_party_skus
        
        self.overlay = data.get('overlay', False)
        self.overlay_compatibility_hook = data.get('overlay_compatibility_hook', False)
        
        aliases = data.get('aliases', None)
        if (aliases is not None) and (not aliases):
            aliases = None
        
        self.aliases = aliases
        
        eula_id = data.get('eula_id', None)
        if eula_id is None:
            eula_id = 0
        else:
            eula_id = int(eula_id)
        
        self.eula_id = eula_id
        
        # Update data may not contain flags, so do not set if missing.
        try:
            flags = data['flags']
        except KeyError:
            pass
        else:
            self.flags = ApplicationFlag(flags)
        
        privacy_policy_url = data.get('privacy_policy_url', None)
        if (privacy_policy_url is not None) and (not privacy_policy_url):
            privacy_policy_url = None
        self.privacy_policy_url = privacy_policy_url
        
        terms_of_service_url = data.get('terms_of_service_url', None)
        if (terms_of_service_url is not None) and (not terms_of_service_url):
            terms_of_service_url = None
        self.terms_of_service_url = terms_of_service_url
        
        self.custom_install_url = data.get('custom_install_url', None)
        
        install_parameters_data = data.get('install_params', None)
        if install_parameters_data is None:
            install_parameters = None
        else:
            install_parameters = ApplicationInstallParameters(data)
        self.install_parameters = install_parameters
        
        tags = data.get('tags', None)
        if (tags is None) or (not tags):
            tags = None
        else:
            tags = tuple(sorted(tags))
        self.tags = tags
        
        # Only set `max_participants` is received
        
        try:
            max_participants = data['max_participants']
        except KeyError:
            pass
        else:
            self.max_participants = max_participants
        
        # Only set `embedded_activity_configuration` if received
        
        try:
            embedded_activity_configuration_data = data['embedded_activity_config']
        except KeyError:
            pass
        else:
            self.embedded_activity_configuration = EmbeddedActivityConfiguration(embedded_activity_configuration_data)
    
    
    @classmethod
    def precreate(cls, application_id, **kwargs):
        """
        Precreates an application with the given parameters.
        
        Parameters
        ----------
        application_id : `int`, `str`
            The application's id.
        **kwargs : keyword parameters
            Additional predefined attributes for the application.
        
        Other parameters
        ----------------
        bot_public : `bool`, Optional (Keyword only)
            Whether not only the application's owner can join the application's bot to guilds.
        bot_require_code_grant : `bool`, Optional (Keyword only)
            Whether the application's bot will only join a guild, when completing the full `oauth2` code grant flow.
        description : `str`, Optional (Keyword only)
            The description of the application.
        flags : `int`, ``ApplicationFlag``, Optional (Keyword only)
            The application's public flags. If not given as ``ApplicationFlag``, then is converted to it.
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The application's icon.
            
            > Mutually exclusive with `icon_type` and `icon_hash`.
        icon_type : ``IconType``, Optional (Keyword only)
            The application's icon's type.
            
            > Mutually exclusive with `icon`.
        icon_hash : `int`, Optional (Keyword only)
            The application's icon's hash.
            
            > Mutually exclusive with `icon`.
        owner : ``ClientUserBase``, ``Team``
            The application's owner. Defaults to `ZEROUSER`.
            
            This field cannot be given as `snowflake`, because it might represent both ``UserBase``-s  and
            ``Team``-s as well.
        privacy_policy_url : `None`, `str`, Optional (Keyword only)
            The url of the application's privacy policy.
        slug : `None`, `str`
            If this application is a game sold on Discord, this field will be the url slug that links to the store page.
        splash : `None`, ``Icon``, `str`, Optional (Keyword only)
            The application's splash.
            
            > Mutually exclusive with `icon_type` and `icon_hash`.
        splash_type : ``IconType``, Optional (Keyword only)
            The application's icon's type.
            
            > Mutually exclusive with `icon`.
        splash_hash : `int`, Optional (Keyword only)
            The application's icon's hash.
            
            > Mutually exclusive with `icon`.
        
        summary : `str`, Optional (Keyword only)
            If this application is a game sold on Discord, this field will be the summary field for the store page of
            its primary sku.
            
            > ``.summary` is deprecated. The parameter will be removed in 2022 Jun.
        
        terms_of_service_url : `None`, `str`, Optional (Keyword only)
            The url of the application's terms of service.
        
        Returns
        -------
        application : `Application``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        application_id = preconvert_snowflake(application_id, 'application_id')
        
        if kwargs:
            processable = []
            
            for key in ('bot_public', 'bot_require_code_grant'):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    continue
                
                value = preconvert_bool(value, key)
                processable.append((key, value))
            
            for key in ('description',):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    continue
                
                value = preconvert_str(value, key, 0, 1024)
                processable.append((key, value))
            
            try:
                kwargs.pop(key)
            except KeyError:
                pass
            else:
                warnings.warn(
                    f'`{cls.__name__}.summary` is deprecated and will be removed in 2022 Jun.',
                    FutureWarning,
                    stacklevel = 2,
                )
            
            try:
                flags = kwargs.pop('flags')
            except KeyError:
                pass
            else:
                flags = preconvert_flag(flags, 'flags', ApplicationFlag)
                processable.append(('flags', flags))
            
            cls.icon.preconvert(kwargs, processable)
            
            try:
                owner = kwargs.pop('owner')
            except KeyError:
                pass
            else:
                if not isinstance(owner, (ClientUserBase, Team)):
                    raise TypeError(
                        f'`owner` can be `{ClientUserBase.__name__}`, `{Team.__name__}`, got '
                        f'{owner.__class__.__name__}; {owner!r}.'
                    )
                
                processable.append(('owner', owner))
            
            for key in ('slug', 'privacy_policy_url', 'terms_of_service_url'):
                try:
                    value = kwargs.pop(key)
                except KeyError:
                    continue
                
                if (value is None):
                    continue
                
                value = preconvert_str(value, key, 0, 1024)
                if not value:
                    continue
                
                processable.append((key, value))
            
            cls.splash.preconvert(kwargs, processable)
            
            if kwargs:
                raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
            
        else:
            processable = None
        
        try:
            application = APPLICATIONS[application_id]
        except KeyError:
            application = cls._create_empty(application_id)
            is_partial = True
        else:
            is_partial = application.partial
        
        if is_partial and (processable is not None):
            for item in processable:
                setattr(application, *item)
        
        return application
    
    
    @property
    def summary(self):
        """
        `.summary` is deprecated and will be removed in 2022 Jun. Please use `.description` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.summary` is deprecated and will be removed in 2022 Jun. '
                f'Please use `.description` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return ''



__all__ = ('CommandCategory',)

from scarletio import RichAttributeErrorBaseType, WeakReferer

from .function import CommandFunction
from .helpers import command_sort_key, normalize_command_name
from .render_constants import BOX_TITLE_SUB_COMMANDS
from .rendering_helpers import render_sub_command_box_into, render_usage_line_into
from .result import (
    COMMAND_RESULT_CODE_CATEGORY_EMPTY, COMMAND_RESULT_CODE_CATEGORY_REQUIRES_PARAMETER,
    COMMAND_RESULT_CODE_CATEGORY_UNKNOWN_SUB_COMMAND, CommandResult
)


class CommandCategory(RichAttributeErrorBaseType):
    """
    Command line command category.
    
    Attributes
    ----------
    _command_categories : `None`, `dict` of (`str`, ``CommandCategory``) items
        Sub commands of the command.
    _command_function : `None`, ``CommandFunction``
        Command to call, if sub command could not be detected.
    _parent_reference : `None`, ``WeakReferer``
        Weakreference to the command category's parent.
    _self_reference : `None`, ``WeakReferer``
        Weakreference to the category itself.
    name : `None`, `str`
        The sub command category's name.
    """
    __slots__ = (
        '__weakref__', '_command_categories', '_command_function', '_parent_reference', '_self_reference', 'name'
    )
    
    def __new__(cls, parent, name):
        """
        Creates a new command line command category.
        
        Parameters
        ----------
        parent : ``Command``, ``CommandCategory``
            The parent command or command category.
        name : `None`, `str`
            The command category's name.
        """
        if (name is not None):
            name = normalize_command_name(name)
        
        self = object.__new__(cls)
        self.name = name
        self._command_function = None
        self._command_categories = None
        self._self_reference = None
        self._parent_reference = parent._self_reference
        
        self._self_reference = WeakReferer(self)
        
        return self
    
    
    def __repr__(self):
        """Returns the command category's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        name = self.name
        if (name is not None):
            repr_parts.append(' name=')
            repr_parts.append(repr(name))
            
            field_added = True
        else:
            field_added = False
        
        command_function = self._command_function
        if (command_function is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' command_function=')
            repr_parts.append(repr(command_function))
        
        sub_categories = [*self.iter_sub_categories()]
        if sub_categories:
            if not field_added:
                repr_parts.append(',')
            
            repr_parts.append(', sub_categories={')
            
            index = 0
            limit = len(sub_categories)
            
            while True:
                sub_category = sub_categories[index]
                
                index += 1
                repr_parts.append(repr(sub_category))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
            
            repr_parts.append('}')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def register_command_category(self, name):
        """
        Registers a sub command to the command.
        
        Parameters
        ----------
        name : `str`
            The name of the sub-command.
        
        Returns
        -------
        sub_command : ``CommandCategory``
        """
        sub_command = CommandCategory(self, name)
        sub_commands = self._command_categories
        if (sub_commands is None):
            sub_commands = {}
            self._command_categories = sub_commands
        
        sub_commands[sub_command.name] = sub_command
        return sub_command
    
    
    def register_command_function(self, function, name, description):
        """
        Parameters
        ----------
        function : `callable`
            The function to call when the command is used.
        name : `str`
            The command's name.
        description : `None`, `str`
            The command's description.
        
        Returns
        -------
        command_function : CommandFunction
        """
        self_name = self.name
        if self_name is None:
            parent_name = self.get_parent_name()
            if parent_name is None:
                direct_register = False
            elif name == parent_name:
                direct_register = True
            else:
                direct_register = False
        
        else:
            if self_name == name:
                direct_register = True
            else:
                direct_register = False
        
        if direct_register:
            command_function = CommandFunction(self, function, name, description)
            self._command_function = command_function
            return command_function
        
        
        category = self.register_command_category(name)
        return category.register_command_function(function, name, description)
    
    
    def get_parent(self):
        """
        Returns the parent category of the command.
        
        Returns
        -------
        parent : `None`, ``Command``, ``CommandCategory``
        """
        parent_reference = self._parent_reference
        if (parent_reference is not None):
            return parent_reference()
    
    
    def get_parent_name(self):
        """
        Returns the category's parent's name.
        
        Returns
        -------
        name : `None`, `str`
        """
        parent = self.get_parent()
        if (parent is not None):
            return parent.name
    
    
    def _trace_back_name(self):
        """
        Traces back to the source name of the command.
        
        This method is an iterable generator.
        
        Yields
        ------
        name : `str`
        
        Returns
        -------
        parent_name : `str`, `None`
        """
        parent_name = None
        
        parent = self.get_parent()
        if (parent is not None):
            parent_name = yield from parent._trace_back_name()
                
        
        name = self.name
        if (name is not None):
            yield name
        
        if (name is None):
            return parent_name
        
        return name
    
    
    def invoke(self, parameters, index):
        """
        Calls the command line command category.
        
        Parameters
        ----------
        parameters : `list` of `str`
            Command line parameters.
        index : `int`
            The index of the first parameter trying to process.
        
        Returns
        -------
        command_result : ``CommandResult``
        """
        command_categories = self._command_categories
        if (command_categories is None):
            command_function = self._command_function
            if (command_function is not None):
                return command_function.invoke(parameters, index)
            
            return CommandResult(
                COMMAND_RESULT_CODE_CATEGORY_EMPTY,
                self,
            )
            
        
        if index >= len(parameters):
            command_function = self._command_function
            if (command_function is not None):
                return command_function.invoke(parameters, index)
            
            return CommandResult(
                COMMAND_RESULT_CODE_CATEGORY_REQUIRES_PARAMETER,
                self,
            )
        
        command_name = parameters[index]
        if command_name.startswith('-'):
            command_category = None
        else:
            command_category = command_categories.get(normalize_command_name(command_name), None)
        
        if (command_category is not None):
            return command_category.invoke(parameters, index + 1)
        
        command_function = self._command_function
        if (command_function is not None):
            return command_function.invoke(parameters, index)
        
        return CommandResult(
            COMMAND_RESULT_CODE_CATEGORY_UNKNOWN_SUB_COMMAND,
            self,
            command_name,
        )
    
    
    def iter_sub_categories(self):
        """
        Iterates over the sub categories of the category.
        
        This method is an iterable generator.
        
        Yields
        ------
        sub_category : ``CommandCategory``
        """
        command_categories = self._command_categories
        if (command_categories is not None):
            yield from command_categories.values()
    
    
    def iter_sub_category_names(self):
        """
        Iterates over the sub-category names of the category.
        
        This method is an iterable generator.
        
        Yields
        ------
        sub_category_name : `str`
        """
        for sub_category in self.iter_sub_categories():
            name = sub_category.name
            if (name is not None):
                yield name
    
    
    def has_sub_categories(self):
        """
        Returns whether the category has sub-categories.
        
        Yields
        ------
        has_sub_categories : `bool`
        """
        return (self._command_categories is not None)
    
    
    def walk_usage(self):
        """
        Walks over the usage of the command category.
        
        This method is an iterable generator.
        
        Yields
        -------
        usage : `str`
        """
        for into in self.walk_usage_into([]):
            yield ''.join(into)
            into.clear()
    
    
    def get_direct_usage(self, *sub_command_stack):
        """
        Returns the direct usage of the command category for the given sub-command stack.
        
        Parameters
        ----------
        *sub_command_stack : `str`
            Sub command stack to get the direct usage for.
        
        Returns
        -------
        usage : `str`
        """
        return ''.join(self.render_direct_usage_into([], *sub_command_stack))
    
    
    def render_direct_usage_into(self, into, *sub_command_stack):
        """
        Renders the direct usage of the command category for the given sub-command stack.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the usage into.
        *sub_command_stack : `str`
            Sub command stack to get the direct usage for.
        
        Returns
        -------
        into : `list` of `str`
        """
        if sub_command_stack:
            command_categories = self._command_categories
            if (command_categories is None):
                category = None
            else:
                category = command_categories.get(normalize_command_name(sub_command_stack[0]), None)
        else:
            category = None
        
        if (category is None):
            command_function = self._command_function
            if (command_function is not None):
                return command_function.render_usage_into(into)
            
            return self.render_partial_usage_into(into)
        
        return category.render_direct_usage_into(into, *sub_command_stack[1:])
    
    
    def walk_usage_into(self, into):
        """
        Walks over the usage of the command category and renders it to the given list.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the usage into.
        
        Yields
        ------
        into : `list` of `str`
        
        Returns
        -------
        into : `list` of `str`
        """
        command_function = self._command_function
        if (command_function is not None):
            into = command_function.render_usage_into(into)
            yield into
        
        for command_category in sorted(self.iter_sub_categories(), key=command_sort_key):
            into = yield from command_category.walk_usage_into(into)
        
        return into
    
    
    def render_partial_usage(self):
        """
        Renders the command category's partial usage.
        
        Returns
        -------
        usage : `str`
        """
        return ''.join(self.render_partial_usage_into([]))
    
    
    def render_partial_usage_into(self, into):
        """
        Renders the command category's partial usage into the given list.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the usage into.
        
        Returns
        -------
        into : `list` of `str`
        """
        into = render_usage_line_into(into, self._trace_back_name())
        into.append(' ...')
        
        sub_command_names = [*self.iter_sub_category_names()]
        sub_command_line_length = max((len(sub_command_name) for sub_command_name in sub_command_names), default=0)
        
        line_length = 0
        if sub_command_line_length:
            line_length = max(sub_command_line_length, len(BOX_TITLE_SUB_COMMANDS))
        
        if sub_command_line_length:
            into.append('\n\n')
            into = render_sub_command_box_into(into, sub_command_names, line_length)
        
        return into
    
    
    def iter_command_functions(self):
        """
        Iterates over the command functions of the command category.
        
        This method is an iterable generator.
        
        Yields
        ------
        command_function : ``CommandFunction``
        """
        command_function = self._command_function
        if (command_function is not None):
            yield command_function
        
        for sub_category in self.iter_sub_categories():
            yield from sub_category.iter_command_functions()



__all__ = ('ComponentTextInput',)

import reprlib

from scarletio import copy_docs, include

from ...preconverters import preconvert_preinstanced_type

from .component_base import ComponentBase
from .debug import (
    _debug_component_custom_id, _debug_component_label, _debug_component_max_length, _debug_component_min_length,
    _debug_component_placeholder, _debug_component_required, _debug_component_text_input_value
)
from .preinstanced import ComponentType, TextInputStyle


create_identifier_custom_id_from_name = include('create_identifier_custom_id_from_name')


class ComponentTextInput(ComponentBase):
    """
    Text input component.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which text input was clicked by the user.
    
    label : `None`, `str`
        Label of the component.
    
    max_length : `int`
        The maximal length of the inputted text.
        
        Defaults to `0` if not applicable.
        
    min_length : `int`
        The minimal length of the inputted text.
        
        Defaults to `0` if not applicable.
    
    placeholder : `str`
        Placeholder text of the text input.
    
    required : `bool`
        Whether the field is required to be fulfilled.
    
    style : `None`, ``TextInputStyle``
        The text input's style.
    
    value : `None`, `str`
        The text input's default value.
    
    Class Attributes
    ----------------
    default_style : ``TextInputStyle`` = `TextInputStyle.short`
        The default text input style to use if style is not given.
    type : ``ComponentType`` = `ComponentType.text_input`
        The component's type.
    """
    default_style = TextInputStyle.short
    type = ComponentType.text_input
    
    __slots__ = (
        'custom_id', 'label', 'max_length', 'min_length', 'placeholder', 'required', 'style', 'value'
    )
    
    def __new__(cls, label=None, *, custom_id=None, max_length=0, min_length=0, placeholder=None, required=None,
            style=None, value=None):
        """
        Creates a new component instance with the given parameters.
        
        Parameters
        ----------
        label : `None`, `str` = `None`, Optional
            Label of the component.
        
        custom_id : `None`, `str` = `None`, Optional (Keyword only)
            Custom identifier to detect which text input was clicked by the user.
            
            If not given, is autogenerated form the `label` parameter.
        
        max_length : `int` = `0`, Optional (Keyword only)
            The maximal length of the inputted text.
            
            Defaults to `0` if not applicable.
        
        min_length : `int` = `0`, Optional (Keyword only)
            The minimal length of the inputted text.
            
            Defaults to `0` if not applicable.
        
        placeholder : `None`, `str` = `None`, Optional (Keyword only)
            Placeholder text of the select.
        
        required : `None`, `bool` = `None`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
            
            If not given, or given as `None`, will default to `True` if `min_length` is defined as higher than `0`.
        
        style : `None`, ``TextInputStyle``, `int` = `None`, Optional (Keyword only)
            The text input's style.
        
        value : `None`, `str` = `None`, Optional (Keyword only)
            The text input's default value.
        
        Raises
        ------
        TypeError
            If `style`'s type is unexpected.
        AssertionError
            - If `custom_id` was not given neither as `None`, `str`.
            - If `style` was not given as any of the `type`'s expected styles.
            - If `label` was not given neither as `None` nor as `int`.
            - If `label`'s length is over `80`.
            - If `custom_id`'s length is over `100`.
            - If `max_length` was not given as `int`.
            - If `man_length`'s is out of the expected range.
            - If `min_length` was not given as `int`.
            - If `min_length`'s is out of the expected range.
            - If `required` is neither `None` nor `bool`.
            - If `value` is neither `None` nor `bool`.
        """
        if __debug__:
            _debug_component_custom_id(custom_id)
            _debug_component_label(label)
            _debug_component_max_length(max_length)
            _debug_component_min_length(min_length)
            _debug_component_placeholder(placeholder)
            _debug_component_required(required)
            _debug_component_text_input_value(value)
        
        # custom_id
        # Only postprocess
        
        # label
        if (label is not None) and (not label):
            label = None
        
        # custom_id | postprocess
        if (custom_id is None) or (not custom_id):
            custom_id = create_identifier_custom_id_from_name(label)
        
        # max_length
        # No additional checks
        
        # min_length
        # No additional checks
        
        # placeholder
        if (placeholder is not None) and (not placeholder):
            placeholder = None
        
        # required
        # If required is `None`, we detect it from `min_value`.
        if (required is None):
            if min_length > 0:
                required = True
            else:
                required = False
        
        # style
        if style is None:
            style = cls.default_style
        else:
            style = preconvert_preinstanced_type(style, 'style', TextInputStyle)

        # value
        if (value is not None) and (not value):
            value = None
        
        self = object.__new__(cls)
        
        self.custom_id = custom_id
        self.label = label
        self.max_length = max_length
        self.min_length = min_length
        self.placeholder = placeholder
        self.required = required
        self.style = style
        self.value = value
        
        return self
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        # custom_id
        self.custom_id = data.get('custom_id', None)
        
        # label
        self.label = data.get('label', None)
        
        # max_length
        self.max_length = data.get('max_length', 0)
        
        # min_length
        self.min_length = data.get('min_length', 0)
        
        # placeholder
        placeholder = data.get('placeholder', None)
        if (placeholder is not None) and (not placeholder):
            placeholder = None
        self.placeholder = placeholder
        
        # required
        self.required = data.get('required', True)
        
        # style
        style = data.get('style', None)
        if (style is not None):
            style = TextInputStyle.get(style)
        self.style = style
        
        # value
        value = data['value']
        if (value is not None) and (not value):
            value = None
        self.value = value
        
        
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        # type
        data = {
            'type': self.type.value,
        }
        
        #  custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            data['custom_id'] = custom_id
        
        # label
        label = self.label
        if (label is not None):
            data['label'] = label
        
        # max_length
        max_length = self.max_length
        if max_length:
            data['max_length'] = max_length
        
        # min_length
        min_length = self.min_length
        if min_length:
            data['min_length'] = min_length
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            data['placeholder'] = placeholder
        
        # required
        if (not self.required):
            data['required'] = False
        
        # style
        style = self.style
        if (style is not None):
            data['style'] = style.value
        
        # value
        value = self.value
        if (value is not None):
            data['value'] = value
        
        return data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type & style
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        # style
        style = self.style
        repr_parts.append(', style=')
        repr_parts.append(style.name)
        repr_parts.append(' (')
        repr_parts.append(repr(style.value))
        repr_parts.append(')')
        
        # System fields : custom_id
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id=')
            repr_parts.append(reprlib.repr(custom_id))
        
        # Text fields : label & placeholder & value
        
        # label
        label = self.label
        if (label is not None):
            repr_parts.append(', label=')
            repr_parts.append(reprlib.repr(label))
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            repr_parts.append(', placeholder=')
            repr_parts.append(repr(placeholder))
        
        # value
        value = self.value
        if (value is not None):
            repr_parts.append(', value=')
            repr_parts.append(repr(value))
        
        # Optional descriptive fields : max_length & min_length & required
        
        # min_length
        min_length = self.min_length
        if min_length:
            repr_parts.append(', min_length=')
            repr_parts.append(repr(min_length))
        
        # min_length
        max_length = self.max_length
        if max_length:
            repr_parts.append(', max_length=')
            repr_parts.append(repr(max_length))
        
        # required (relation with `min_length`)
        required = self.required
        if (min_length > 0) ^ required:
            repr_parts.append(', required=')
            repr_parts.append(repr(required))
        
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.custom_id = self.custom_id
        new.label = self.label
        new.max_length = self.max_length
        new.min_length = self.min_length
        new.placeholder = self.placeholder
        new.required = self.required
        new.style = self.style
        new.value = self.value
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        
        Other Parameters
        ----------------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which text input was clicked by the user.
        
        label : `None`, `str`, Optional (Keyword only)
            Label of the component.
        
        max_length : `int`, Optional (Keyword only)
            The maximal length of the inputted text.
        
        min_length : `int`, Optional (Keyword only)
            The minimal length of the inputted text.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder text of the select.
        
        required : `None`, `bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
            
            If not given, or given as `None`, will default to `True` if `min_length` is defined as higher than `0`.
        
        style : `None`, ``TextInputStyle``, `int`, Optional (Keyword only)
            The text input's style.
        
        value : `None`, `str`, Optional (Keyword only)
            The text input's default value.
        
        Returns
        -------
        new : ``ComponentTextInput``
        """
        # custom_id
        try:
            custom_id = kwargs.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            if __debug__:
                _debug_component_custom_id(custom_id)
            
            if (custom_id is not None) and (not custom_id):
                custom_id = None
        
        # label
        try:
            label = kwargs.pop('label')
        except KeyError:
            label = self.label
        else:
            if __debug__:
                _debug_component_label(label)
        
        # custom_id | postprocess
        if (custom_id is None):
            custom_id = create_identifier_custom_id_from_name(label)
        
        
        # max_length
        try:
            max_length = kwargs.pop('max_length')
        except KeyError:
            max_length = self.max_length
        else:
            if __debug__:
                _debug_component_max_length(max_length)
        
        # min_length
        try:
            min_length = kwargs.pop('min_length')
        except KeyError:
            min_length = self.min_length
        else:
            if __debug__:
                _debug_component_min_length(min_length)
        
        # placeholder
        try:
            placeholder = kwargs.pop('placeholder')
        except KeyError:
            placeholder = self.placeholder
        else:
            if __debug__:
                _debug_component_placeholder(placeholder)
            
            if (placeholder is not None) and (not placeholder):
                placeholder = None
        
        # required
        try:
            required = kwargs.pop('required')
        except KeyError:
            required = self.required
        else:
            if __debug__:
                _debug_component_required(required)
            
            if (required is None):
                if min_length > 0:
                    required = True
                else:
                    required = False
        
        # style
        try:
            style = kwargs.pop('style')
        except KeyError:
            style = self.style
        
        if style is None:
            style = self.default_style
        else:
            style = preconvert_preinstanced_type(style, 'style', TextInputStyle)
        
        # value
        try:
            value = kwargs.pop('value')
        except KeyError:
            value = None
        else:
            if __debug__:
                _debug_component_text_input_value(value)
            
            if (value is not None) and (not value):
                value = None
        
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
        
        new = object.__new__(type(self))
        
        new.custom_id = custom_id
        new.label = label
        new.max_length = max_length
        new.min_length = min_length
        new.placeholder = placeholder
        new.required = required
        new.style = style
        new.value = value
        
        return new
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # label
        if self.label != other.label:
            return False
        
        # max_length
        if self.max_length != other.max_length:
            return False
        
        # min_length
        if self.min_length != other.min_length:
            return False
        
        # placeholder
        if self.placeholder != other.placeholder:
            return False
        
        # required
        if self.required != other.required:
            return False
        
        # style
        if self.style is not other.style:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        # type
        hash_value = self.type.value
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= hash(label)
        
        # max_length
        max_length = self.max_length
        if max_length:
            hash_value ^= max_length << 12
        
        # min_length
        min_length = self.min_length
        if min_length:
            hash_value ^= min_length << 20
        
        # placeholder
        placeholder = self.placeholder
        if (placeholder is not None):
            hash_value ^= hash(placeholder)
        
        # required
        if self.required:
            hash_value ^= (1 << 28)
        
        # style
        style = self.style
        if (style is not None):
            hash_value ^= style.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value



__all__ = ()

from importlib.machinery import ModuleSpec

from scarletio import export

from .source_loader import PluginSourceLoader


@export
class PluginModuleSpecType(ModuleSpec):
    """
    Extends the builtin spec with plugin checking.
    """
    def __init__(self, other):
        """
        `ModuleSpec` subclass implementing rich context for plugins.
        
        Parameters
        ----------
        other : ``ModuleSpec``
            The other module spec to inherit from.
        """
        self.__dict__.update(other.__dict__)
        loader = other.loader
        self.loader = PluginSourceLoader(loader.name, loader.path)
        self._initializing_internal = False
    
    
    def is_initialised(self):
        """
        Returns whether the module spec is initialised.
        
        Returns
        -------
        is_initialised : `bool`
        """
        loader = self.loader
        if loader is None:
            return False
        
        if not isinstance(loader, PluginSourceLoader):
            return False
        
        if loader._module is None:
            return False
        
        return True
    
    
    def get_module(self):
        """
        Returns the module spec's real module.
        
        Returns
        -------
        module : `None`, ``ModuleType``
        """
        loader = self.loader
        if loader is None:
            return None
        
        if not isinstance(loader, PluginSourceLoader):
            return None
        
        return loader._module
    
    
    def get_module_proxy(self):
        """
        Returns the module spec's proxy module.
        
        Returns
        -------
        module : `None`, ``PluginModuleProxyType``
        """
        loader = self.loader
        if loader is None:
            return None
        
        if not isinstance(loader, PluginSourceLoader):
            return None
        
        return loader._module_proxy



__all__ = ('ComponentButton', )

import reprlib

from scarletio import copy_docs, include

from ...emoji import create_partial_emoji_data, create_partial_emoji_from_data
from ...preconverters import preconvert_preinstanced_type
from ...utils import url_cutter

from .component_base import ComponentBase
from .debug import (
    _debug_component_custom_id, _debug_component_emoji, _debug_component_enabled, _debug_component_label,
    _debug_component_url
)
from .preinstanced import ButtonStyle, ComponentType


create_auto_custom_id = include('create_auto_custom_id')


class ComponentButton(ComponentBase):
    """
    Button component.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which button was clicked by the user.
        
        > Mutually exclusive with the `url` field.
    
    enabled : `bool`
        Whether the component is enabled.
    
    emoji : `None`, ``Emoji``
        Emoji of the button if applicable.
    
    label : `None`, `str`
        Label of the component.
    
    style : `None`, ``ButtonStyle``
        The button's style.
    
    url : `None`, `str`
        Url to redirect to when clicking on the button.
        
        > Mutually exclusive with the `custom_id` field.
    
    Class Attributes
    ----------------
    default_style : ``ButtonStyle`` = `ButtonStyle.blue`
        The default button style to use if style is not given.
    type : ``ComponentType`` = `ComponentType.button`
        The component's type.
    """
    default_style = ButtonStyle.blue
    type = ComponentType.button
    
    __slots__ = ('custom_id', 'enabled', 'emoji', 'label', 'style', 'url',)
    
    def __new__(cls, label=None, emoji=None, *, custom_id=None, enabled=True, style=None, url=None):
        """
        Creates a new component instance with the given parameters.
        
        Parameters
        ----------
        label : `None`, `str` = `None`, Optional
            Label of the component.
        
        emoji : `None`, ``Emoji`` = `None`, Optional
            Emoji of the button if applicable.
        
        custom_id : `None`, `str` = `None`, Optional (Keyword only)
            Custom identifier to detect which button was clicked by the user.
            
            > Mutually exclusive with the `url` field.
        
        enabled : `bool` = `True`, Optional (Keyword only)
            Whether the button is enabled. Defaults to `True`.
        
        style : `None`, ``ButtonStyle``, `int` = `None`, Optional (Keyword only)
            The button's style.
        
        url : `None`, `str` = `None`, Optional (Keyword only)
            Url to redirect to when clicking on the button.
            
            > Mutually exclusive with the `custom_id` field.
        
        Raises
        ------
        TypeError
            If `style`'s type is unexpected.
        AssertionError
            - If `custom_id` was not given neither as `None`, `str`.
            - `url` is mutually exclusive with `custom_id`.
            - If `emoji` was not given as ``Emoji``.
            - If `url` was not given neither as `None`, `str`.
            - If `style` was not given as any of the `type`'s expected styles.
            - If `label` was not given neither as `None` nor as `int`.
            - If `enabled` was not given as `bool`.
            - If `label`'s length is over `80`.
            - If `custom_id`'s length is over `100`.
        """
        if __debug__:
            _debug_component_custom_id(custom_id)
            _debug_component_enabled(enabled)
            _debug_component_emoji(emoji)
            _debug_component_label(label)
            _debug_component_url(url)
        
        # custom_id
        if (custom_id is not None) and (not custom_id):
            custom_id = None
        
        # enabled
        # No additional checks
        
        # emoji
        # No additional checks
        
        # label
        if (label is not None) and (not label):
            label = None
        
        # url
        if (url is not None) and (not url):
            url = None
        
        # custom_id & custom_id mixed
        if __debug__:
            if (custom_id is not None) and (url is not None):
                raise AssertionError(
                    f'`custom_id` and `url` fields are mutually exclusive, got '
                    f'custom_id={custom_id!r}, url={url!r}.'
                )
        
        # style
        if (url is None):
            # style
            if style is None:
                style = cls.default_style
            else:
                style = preconvert_preinstanced_type(style, 'style', ButtonStyle)
            
            # If `custom_id` is required, but not given, generate one.
            if (custom_id is None):
                custom_id = create_auto_custom_id()
        
        else:
            style = ButtonStyle.link
        
        
        self = object.__new__(cls)
        
        self.custom_id = custom_id
        self.enabled = enabled
        self.emoji = emoji
        self.label = label
        self.style = style
        self.url = url
        
        return self
    
    
    @classmethod
    @copy_docs(ComponentBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        # custom_id
        self.custom_id = data.get('custom_id', None)
        
        # enabled
        self.enabled = not data.get('disabled', False)
        
        # emoji
        emoji_data = data.get('emoji', None)
        if emoji_data is None:
            emoji = None
        else:
            emoji = create_partial_emoji_from_data(emoji_data)
        self.emoji = emoji
        
        # label
        self.label = data.get('label', None)
        
        # style
        self.style = ButtonStyle.get(data.get('style', 0))
        
        # url
        self.url = data.get('url', None)
        
        return self
    
    
    @copy_docs(ComponentBase.to_data)
    def to_data(self):
        # type
        data = {
            'type' : self.type.value
        }
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            data['custom_id'] = custom_id
        
        # enabled
        if (not self.enabled):
            data['disabled'] = True
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            data['emoji'] = create_partial_emoji_data(emoji)
        
        # label
        label = self.label
        if (label is not None):
            data['label'] = label
        
        # style
        style = self.style
        if (style is not None):
            data['style'] = style.value
        
        # url
        url = self.url
        if (url is not None):
            data['url'] = url
        
        return data
    
    
    @copy_docs(ComponentBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__, ]
        
        # Descriptive fields : type & style
        
        # type
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append(' (')
        repr_parts.append(repr(type_.value))
        repr_parts.append(')')
        
        # style
        style = self.style
        repr_parts.append(', style=')
        repr_parts.append(style.name)
        repr_parts.append(' (')
        repr_parts.append(repr(style.value))
        repr_parts.append(')')
        
        # System fields : custom_id
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id=')
            repr_parts.append(reprlib.repr(custom_id))
        
        # Text fields : emoji & label
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji=')
            repr_parts.append(repr(emoji))
        
        # label
        label = self.label
        if (label is not None):
            repr_parts.append(', label=')
            repr_parts.append(reprlib.repr(label))
        
        
        # Optional descriptive fields: url & enabled
        
        # url
        url = self.url
        if (url is not None):
            repr_parts.append(', url=')
            repr_parts.append(url_cutter(url))
        
        # enabled
        enabled = self.enabled
        if (not enabled):
            repr_parts.append(', enabled=')
            repr_parts.append(repr(enabled))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.custom_id = self.custom_id
        new.enabled = self.enabled
        new.emoji = self.emoji
        new.label = self.label
        new.style = self.style
        new.url = self.url
        
        return new
    
    
    def copy_with(self, **kwargs):
        """
        Copies the component and modifies the created one with the given parameters.
        
        Parameters
        ----------
        **kwargs : Keyword parameters
            Keyword parameters referencing attributes.
        
        Other Parameters
        ----------------
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which button was clicked by the user.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the button is enabled. Defaults to `True`.
        
        emoji : `None`, ``Emoji``, Optional (Keyword only)
            Emoji of the button if applicable.
        
        label : `None`, `str`, Optional (Keyword only)
            Label of the component.
        
        style : `None`, ``ButtonStyle``, `int`, Optional (Keyword only)
            The button's style.
        
        url : `None`, `str`, Optional (Keyword only)
            Url to redirect to when clicking on the button.
        
        Returns
        -------
        new : ``ComponentButton``
        """
        # custom_id
        try:
            custom_id = kwargs.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            if __debug__:
                _debug_component_custom_id(custom_id)
            
            if (custom_id is not None) and (not custom_id):
                custom_id = None
        
        # enabled
        try:
            enabled = kwargs.pop('enabled')
        except KeyError:
            enabled = self.enabled
        else:
            if __debug__:
                _debug_component_enabled(enabled)
        
        # emoji
        try:
            emoji = kwargs.pop('emoji')
        except KeyError:
            emoji = self.emoji
        else:
            if __debug__:
                _debug_component_emoji(emoji)
        
        # label
        try:
            label = kwargs.pop('label')
        except KeyError:
            label = self.label
        else:
            if __debug__:
                _debug_component_label(label)
        
        # style
        try:
            style = kwargs.pop('style')
        except KeyError:
            style = self.style
        
        # url
        try:
            url = kwargs.pop('url')
        except KeyError:
            url = self.url
        else:
            if __debug__:
                _debug_component_url(url)
            
            if (url is not None) and (not url):
                url = None

        
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: {kwargs!r}.')
        
        
        # custom_id & url
        if __debug__:
            if (custom_id is not None) and (url is not None):
                raise AssertionError(
                    f'`custom_id` and `url` fields are mutually exclusive, got '
                    f'custom_id={custom_id!r}, url={url!r}.'
                )
        
        # url # style & custom
        if (url is None):
            if style is None:
                style = self.default_style
            else:
                style = preconvert_preinstanced_type(style, 'style', ButtonStyle)
            
            if (custom_id is None):
                custom_id = create_auto_custom_id()
        
        else:
            style = ButtonStyle.link
        
        new = object.__new__(type(self))
        
        new.custom_id = custom_id
        new.enabled = enabled
        new.emoji = emoji
        new.label = label
        new.url = url
        new.style = style
        
        return new
    
    
    @copy_docs(ComponentBase.__eq__)
    def __eq__(self, other):
        if type(other) is not type(self):
            return NotImplemented
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # label
        if self.label != other.label:
            return False
        
        # style
        if self.style is not other.style:
            return False
            
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    @copy_docs(ComponentBase.__hash__)
    def __hash__(self):
        hash_value = self.type.value
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # enabled
        if self.enabled:
            hash_value ^= 1 << 8
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= emoji.id
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= hash(label)
        
        # style
        style = self.style
        if (style is not None):
            hash_value ^= style.value
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= hash(url)
        
        return hash_value



__all__ = ('AutoModerationRule',)

from ..bases import DiscordEntity
from ..channel import Channel, create_partial_channel_from_id
from ..core import AUTO_MODERATION_RULES
from ..preconverters import preconvert_preinstanced_type, preconvert_str
from ..role import Role, create_partial_role_from_id
from ..user import ZEROUSER, create_partial_user_from_id

from .action import AutoModerationAction
from .preinstanced import AutoModerationEventType, AutoModerationRuleTriggerType


def _validate_actions(actions):
    """
    Validates the given `actions` parameter.
    
    Parameters
    ----------
    actions : `None`, (`list`, `set`, `tuple`) of ``AutoModerationAction``
        Actions which will execute when the rule is triggered.
    
    Returns
    -------
    processed_actions : `None`, `tuple` of ``AutoModerationAction``
    
    Raises
    ------
    TypeError
        - If `actions`'s type is incorrect.
        - If an action's type is incorrect.
    """
    if actions is None:
        processed_actions = None
    
    elif isinstance(actions, AutoModerationAction):
        processed_actions = (actions, )
    
    else:
        iterator = getattr(type(actions), '__iter__', None)
        if iterator is None:
            raise TypeError(
                f'`actions` can be `None`, `{AutoModerationAction.__name__}`, '
                f'`iterable` of `{AutoModerationAction.__name__}`, '
                f'got {actions.__class__.__name__}; {actions!r}.'
            )
        
        processed_actions = None
        
        for action in actions:
            if not isinstance(action, AutoModerationAction):
                raise TypeError(
                    f'`actions` can can contain `{AutoModerationAction.__name__}` elements, '
                    f'got {action.__class__.__name__}; {action!r}; actions={actions!r}.'
                )
            
            if processed_actions is None:
                processed_actions = set()
            
            processed_actions.add(action)
    
    if (processed_actions is not None):
        processed_actions = tuple(sorted(processed_actions))
    
    return processed_actions


def _validate_enabled(enabled):
    """
    Validates the given `enabled` parameter.
    
    Parameters
    ----------
    enabled : `bool`
        Whether the rule is enabled.
    
    Returns
    -------
    enabled : `bool`
    
    Raises
    ------
    TypeError
        - If `enabled`'s type is incorrect.
    """
    if not isinstance(enabled, bool):
        raise TypeError(
            f'`enabled` can be `bool`, got {enabled.__class__.__name__}; {enabled!r}.'
        )
    
    return enabled


def _validate_event_type(event_type):
    """
    Validates the given `event_type` parameter.
    
    Parameters
    ----------
    event_type : `None`, `int`, ``AutoModerationEventType``
        For which events is the rule applied.
    
    Returns
    -------
    event_type : ``AutoModerationEventType``
    
    Raises
    ------
    TypeError
        - If `event_type`'s type is incorrect.
    """
    if event_type is None:
        event_type = AutoModerationEventType.message_send
    
    else:
        event_type = preconvert_preinstanced_type(event_type, 'event_type', AutoModerationEventType)
    
    return event_type


def _validate_excluded_channel_ids(excluded_channels):
    """
    Validates the given `excluded_channels` parameter.
    
    Parameters
    ----------
    excluded_channels : `None`, `int`, `Channel`, `iterable` of (`int`, ``Channel``)
        Excluded channels from the rule.
    
    Returns
    -------
    excluded_channel_ids : `None`, `tuple` of `int`
    
    Raises
    ------
    TypeError
        - If `excluded_channels`'s type is incorrect.
        - If an excluded channel's type is incorrect.
    """
    if excluded_channels is None:
        excluded_channel_ids = None
    
    elif isinstance(excluded_channels, int):
        excluded_channel_ids = (excluded_channels, )
    
    elif isinstance(excluded_channels, Channel):
        excluded_channel_ids = (excluded_channels.id, )
    
    else:
        iterator = getattr(type(excluded_channels), '__iter__', None)
        if iterator is None:
            raise TypeError(
                f'`excluded_channels` can be `None`, `int`, `{Channel.__name__}`, '
                f'`iterable` of (`int`, `{Channel.__name__}`), '
                f'got {excluded_channels.__class__.__name__}; {excluded_channels!r}.'
            )
        
        excluded_channel_ids = None
        
        for excluded_channel in excluded_channels:
            if isinstance(excluded_channel, Channel):
                excluded_channel_id = excluded_channel.id
            
            elif isinstance(excluded_channel, int):
                excluded_channel_id = excluded_channel
            
            else:
                raise TypeError(
                    f'`excluded_channels` can can contain (`int`, `{Channel.__name__}`) elements, '
                    f'got {excluded_channel.__class__.__name__}; {excluded_channel!r}; '
                    f'actions={excluded_channels!r}.'
                )
            
            if excluded_channel_ids is None:
                excluded_channel_ids = set()
            
            excluded_channel_ids.add(excluded_channel_id)
        
        if (excluded_channel_ids is not None):
            excluded_channel_ids = tuple(sorted(excluded_channel_ids))
    
    return excluded_channel_ids


def _validate_excluded_role_ids(excluded_roles):
    """
    Validates the given `excluded_roles` parameter.
    
    Parameters
    ----------
    excluded_roles : `None`, `int`, ``Role``, `iterable` of (`int`, ``Role``)
        Excluded roles from the rule.
    
    Returns
    -------
    excluded_role_ids : `None`, `tuple` of `int`
    
    Raises
    ------
    TypeError
        - If `excluded_roles`'s type is incorrect.
        - If an excluded role's type is incorrect.
    """
    if excluded_roles is None:
        excluded_role_ids = None
    
    elif isinstance(excluded_roles, int):
        excluded_role_ids = (excluded_roles, )
    
    elif isinstance(excluded_roles, Role):
        excluded_role_ids = (excluded_roles.id, )
    
    else:
        iterator = getattr(type(excluded_roles), '__iter__', None)
        if iterator is None:
            raise TypeError(
                f'`excluded_roles` can be `None`, `int`, `{Role.__name__}`, '
                f'`iterable` of (`int`, `{Role.__name__}`), '
                f'got {excluded_roles.__class__.__name__}; {excluded_roles!r}.'
            )
        
        excluded_role_ids = None
        
        for excluded_role in excluded_roles:
            if isinstance(excluded_role, Role):
                excluded_role_id = excluded_role.id
            
            elif isinstance(excluded_role, int):
                excluded_role_id = excluded_role
            
            else:
                raise TypeError(
                    f'`excluded_roles` can can contain (`int`, `{Role.__name__}`) elements, '
                    f'got {excluded_role.__class__.__name__}; {excluded_role!r}; '
                    f'actions={excluded_roles!r}.'
                )
            
            if excluded_role_ids is None:
                excluded_role_ids = set()
            
            excluded_role_ids.add(excluded_role_id)
        
        if (excluded_role_ids is not None):
            excluded_role_ids = tuple(sorted(excluded_role_ids))
    
    return excluded_role_ids


def _validate_name(name):
    """
    Validates the given `name` parameter.
    
    Parameters
    ----------
    name : `str`
        The rule's name.
    
    Returns
    -------
    name : `str`
    
    Raises
    ------
    TypeError
        - If `enabled`'s type is incorrect.
    ValueError
        - If `name`'s length is out of the expected range.
    """
    return preconvert_str(name, 'name', 1, 2048)


def _validate_trigger_type_with_metadata_options(
    trigger_type, excluded_keywords, keyword_presets, keywords, mention_limit
):
    """
    Validates the given `trigger_type` with the `keyword_presets`, `keywords` options. If any option is given, the
    `trigger_type` will default towards it. On any mismatch exception is raised.
    
    Parameters
    ----------
    trigger_type : `Ellipsis`, `int`, ``AutoModerationRuleTriggerType``
        Auto moderation trigger type.
    
    keyword_presets : `Ellipsis`, `None`, `int`, ``AutoModerationKeywordPresetType``, \
            `iterable` of (`int`, ``AutoModerationKeywordPresetType``)
        Keyword preset defined by Discord which will be searched for in content.
    
    excluded_keywords : `Ellipsis`, `None`, `str`, `iterable` of `str`
        Excluded keywords from preset filter.

    keywords : `Ellipsis`, `None`, `str`, `iterable` of `str`
        Substrings which will be searched for in content.
    
    mention_limit : `Ellipsis`, `None`, `int`
        The amount of mentions in a message after the rule is triggered.
    
    Returns
    -------
    trigger_metadata : `None`, ``AutoModerationRuleTriggerMetadata``
        Trigger type specific metadata if applicable.
    trigger_type : ``AutoModerationRuleTriggerType``
        The final processed trigger type.
    
    Raises
    ------
    TypeError
        - If a parameter's type is incorrect.
        - If there are multiple mutually exclusive options.
        - If trigger type not detectable.
        - If multiple trigger types detected.
    ValueError
        - If a parameter's value is incorrect.
    """
    if (trigger_type is not ...):
        trigger_type = preconvert_preinstanced_type(trigger_type, 'trigger_type', AutoModerationRuleTriggerType)
    
    if (keyword_presets is not ...) + (keywords is not ...) + (mention_limit is not ...)> 1:
        raise TypeError(
            f'`keyword_presets` and `keywords` parameters are mutually exclusive, got '
            f'keyword_presets={keyword_presets!r}; keywords={keywords!r}.'
        )
    
    if (keyword_presets is not ...):
        if (excluded_keywords is ...):
            excluded_keywords = None
        
        probable_trigger_type = AutoModerationRuleTriggerType.keyword_preset
        metadata_parameters = (keyword_presets, excluded_keywords, )
    
    elif (excluded_keywords is not ...):
        raise TypeError(
            f'`excluded_keywords` is only meaningful with `keyword_presets` parameter, '
            f'got excluded_keywords={excluded_keywords!r}.'
        )
    
    elif (keywords is not ...):
        probable_trigger_type = AutoModerationRuleTriggerType.keyword
        metadata_parameters = (keywords, )
    
    elif (mention_limit is not ...):
        probable_trigger_type = AutoModerationRuleTriggerType.mention_spam
        metadata_parameters = (mention_limit, )
        
    else:
        probable_trigger_type = None
        metadata_parameters = ()
    
    
    if (trigger_type is ...):
        if (probable_trigger_type is None):
            raise TypeError(
                f'`trigger_type` is not given or is given as `None`, and cannot be detected from `keyword_presets` '
                'or from `keywords` parameters.'
            )
        
        trigger_type = probable_trigger_type
    
    else:
        if (probable_trigger_type is not None):
            if (trigger_type is not probable_trigger_type):
                if (keyword_presets is not ...):
                    received_parameter_name = 'keyword_presets'
                    received_parameter_value = keyword_presets
                
                elif (keywords is not ...):
                    received_parameter_name = 'keywords'
                    received_parameter_value = keywords
                
                else:
                    received_parameter_name = 'mention_limit'
                    received_parameter_value = mention_limit
                
                raise TypeError(
                    f'Both `trigger_type` and `{received_parameter_name}` parameters refer to a different '
                    f'trigger type, got trigger_type={trigger_type!r}, '
                    f'{received_parameter_name}={received_parameter_value!r}.'
                )
    
    trigger_metadata_type = trigger_type.metadata_type
    if (trigger_metadata_type is None):
        trigger_metadata = None
    
    else:
        trigger_metadata = trigger_metadata_type(*metadata_parameters)
    
    return trigger_metadata, trigger_type


class AutoModerationRule(DiscordEntity, immortal=True):
    """
    Auto moderation feature which allows guilds to set rules that trigger based on some criteria.
    
    Attributes
    ----------
    id : `int`
        The unique identifier of the auto moderation rule. Defaults to `0`.
    
    actions : `None`, `tuple` of ``AutoModerationAction``
        Actions which will execute when the rule is triggered. Defaults to `0`.
    
    creator_id : `int`
        The user who created the rule. Defaults to `0`.
    
    enabled : `bool`
        Whether the rule is enabled.
    
    event_type : ``AutoModerationEventType``
        For which events is the rule applied.
    
    excluded_channel_ids : `None`, `tuple` of `int`
        The excluded channels' identifiers.
    
    excluded_role_ids : `None`, `tuple` of `int`
        The excluded roles' identifiers.
    
    guild_id : `int`
        The guild's identifier where the role is. Defaults to `0`.
    
    name : `str`
        The rule's name.
    
    trigger_metadata : `None`, ``AutoModerationRuleTriggerMetadata``
        Trigger type specific metadata if applicable.
    
    trigger_type : ``AutoModerationRuleTriggerType``
        Characterizes the type of content which can trigger the rule.
    """
    __slots__ = (
        'actions', 'creator_id', 'enabled', 'event_type', 'excluded_channel_ids', 'excluded_role_ids', 'guild_id',
        'name', 'trigger_metadata', 'trigger_type'
    )
    
    
    def __new__(
        cls, name, actions, trigger_type=..., *, enabled=True, event_type=None, excluded_channels=None,
        excluded_keywords=..., excluded_roles=None, keyword_presets=..., keywords=..., mention_limit=...
    ):
        """
        Creates a new auto moderation rule with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The rule's name.
        
        actions : `None`, ``AutoModerationAction``, `iterable` of ``AutoModerationAction``
             Actions which will execute when the rule is triggered.
        
        trigger_type : ``AutoModerationRuleTriggerType``, `int`, Optional
            Auto moderation trigger type.
            
            By passing `keyword_presets` parameter you can define the trigger type as 
            `AutoModerationRuleTriggerType.keyword_preset`, or by passing the `keywords` you can define it as 
            `AutoModerationRuleTriggerType.keyword`.
            
            > `keyword_preset` triggers also accept an additional `excluded_keywords` parameter.
        
        enabled : `bool` = `True`, Optional (Keyword only)
            Whether the rule is enabled.
        
        event_type : `None`, `int`, ``AutoModerationEventType`` = `AutoModerationEventType.message_send` \
                , Optional (Keyword only)
            For which events is the rule applied.
        
        excluded_channels : `None`, `int`, ``Channel``, `iterable` of (`int`, ``Channel``) = `None`, \
                Optional (Keyword only)
            Excluded channels from the rule.
        
        excluded_keywords : `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            Excluded keywords from preset filter.
            
            > Only meaningful with the `keyword_presets` parameter.
        
        excluded_roles : `None`, `int`, ``Role``, `iterable` of (`int`, ``Role``) = `None`, Optional (Keyword only)
            Excluded roles from the rule.
        
        keyword_presets : `None`, `int`, ``AutoModerationKeywordPresetType``, \
                `iterable` of (`int`, ``AutoModerationKeywordPresetType``), Optional (Keyword only)
            Keyword preset defined by Discord which will be searched for in content.
            
            > Mutually exclusive with the `keywords` and `mention_limit` parameters.
        
        keywords : `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            Substrings which will be searched for in content.
            
            > Mutually exclusive with the `keyword_presets` and `mention_limit` parameters.
        
        mention_limit : `None`, `int`, Optional (Keyword only)
            The amount of mentions in a message after the rule is triggered.
            
            > Mutually exclusive with the `keyword_presets` and `keywords` parameters.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # actions
        actions = _validate_actions(actions)
        
        # enabled
        enabled = _validate_enabled(enabled)
        
        # event_type
        event_type = _validate_event_type(event_type)
        
        # excluded_channel_ids
        excluded_channel_ids = _validate_excluded_channel_ids(excluded_channels)
        
        # excluded_role_ids
        excluded_role_ids = _validate_excluded_role_ids(excluded_roles)
        
        # name
        name = _validate_name(name)
        
        # trigger_type & keywords & keyword_presets
        trigger_metadata, trigger_type = _validate_trigger_type_with_metadata_options(
            trigger_type, excluded_keywords, keyword_presets, keywords, mention_limit
        )
        
        self = object.__new__(cls)
        self.id = 0
        self.actions = actions
        self.creator_id = 0
        self.enabled = enabled
        self.event_type = event_type
        self.excluded_channel_ids = excluded_channel_ids
        self.excluded_role_ids = excluded_role_ids
        self.guild_id = 0
        self.name = name
        self.trigger_metadata = trigger_metadata
        self.trigger_type = trigger_type
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``AutoModerationRule`` from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received auto moderation rule data.
        
        Returns
        -------
        self : ``AutoModerationRule``
            The created auto moderation rule.
        """
        auto_moderation_rule_id = int(data['id'])
        
        try:
            self = AUTO_MODERATION_RULES[auto_moderation_rule_id]
        except KeyError:
            self = object.__new__(cls)
            
            self.id = auto_moderation_rule_id
            
            # creator_id
            creator_id = data.get('creator_id', None)
            if creator_id is None:
                creator_id = 0
            else:
                creator_id = int(creator_id)
            self.creator_id = creator_id
            
            # guild_id
            guild_id = data.get('guild_id', None)
            if (guild_id is None):
                guild_id = 0
            else:
                guild_id = int(guild_id)
            self.guild_id = guild_id
            
            AUTO_MODERATION_RULES[auto_moderation_rule_id] = self
        
        self._update_attributes(data)
        return self
    
    
    def to_data(self):
        """
        Converts the auto moderation rule to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # id
        # Internal field
        
        # actions
        data['actions'] = [auto_moderation_action.to_data() for auto_moderation_action in self.iter_actions()]
        
        # creator_id
        # Internal field
        
        # enabled
        data['enabled'] = self.enabled
        
        # event_type
        data['event_type'] = self.event_type.value
        
        # excluded_channel_ids
        data['exempt_channels'] = [*self.iter_excluded_channel_ids()]
        
        # excluded_role_ids
        data['exempt_roles'] = [*self.iter_excluded_role_ids()]
        
        # guild_id
        # Internal field
        
        # name
        data['name'] = self.name
        
        # trigger_metadata
        trigger_metadata = self.trigger_metadata
        if (trigger_metadata is None):
            trigger_metadata_data = {}
        else:
            trigger_metadata_data = trigger_metadata.to_data()
        data['trigger_metadata'] = trigger_metadata_data
        
        # trigger_type
        data['trigger_type'] = self.trigger_type.value
        
        return data
    
    
    def _update_attributes(self, data):
        """
        Updates the auto moderation rule with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received guild profile data.
        """
        # id
        # Already set & cannot change
        
        # actions
        action_array = data['actions']
        if action_array:
            actions = tuple(
                AutoModerationAction.from_data(auto_moderation_action_data)
                for auto_moderation_action_data in action_array
            )
        else:
            actions = None
        
        self.actions = actions
        
        # creator_id
        # Internal field & cannot change
        
        # enabled
        self.enabled = data['enabled']
        
        # event_type
        self.event_type = AutoModerationEventType.get(data['event_type'])
        
        # excluded_channel_ids
        excluded_channel_id_array = data['exempt_channels']
        if excluded_channel_id_array:
            excluded_channel_ids = tuple(sorted(int(channel_id) for channel_id in excluded_channel_id_array))
        else:
            excluded_channel_ids = None
        self.excluded_channel_ids = excluded_channel_ids
        
        # excluded_role_ids
        excluded_role_id_array = data['exempt_roles']
        if excluded_role_id_array:
            excluded_role_ids = tuple(sorted(int(role_id) for role_id in excluded_role_id_array))
        else:
            excluded_role_ids = None
        self.excluded_role_ids = excluded_role_ids
        
        # guild_id
        # Internal field & cannot change
        
        # name
        self.name = data['name']
        
        # trigger_metadata & trigger_type
        trigger_type = AutoModerationRuleTriggerType.get(data['trigger_type'])
        trigger_metadata_type = trigger_type.metadata_type
        if (trigger_metadata_type is None):
            trigger_metadata = None
        else:
            trigger_metadata = trigger_metadata_type.from_data(data['trigger_metadata'])
        
        self.trigger_type = trigger_type
        self.trigger_metadata = trigger_metadata
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the auto moderation rule and returns the changed attributes in a dictionary with the changed attributes
        in a `attribute-name`, `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received guild profile data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
            
            Might contain the following items:
            
            +-------------------------------+-----------------------------------------------------------+
            | Keys                          | Values                                                    |
            +===============================+===========================================================+
            | actions                       | `None`, `tuple` of ``AutoModerationAction``               |
            +-------------------------------+-----------------------------------------------------------+
            | enabled                       | `bool`                                                    |
            +-------------------------------+-----------------------------------------------------------+
            | event_type                    | ``AutoModerationEventType``                               |
            +-------------------------------+-----------------------------------------------------------+
            | excluded_channel_ids          | `None`, `tuple` of `int`                                  |
            +-------------------------------+-----------------------------------------------------------+
            | excluded_role_ids             | `None`, `tuple` of `int`                                  |
            +-------------------------------+-----------------------------------------------------------+
            | name                          | `str`                                                     |
            +-------------------------------+-----------------------------------------------------------+
            | trigger_metadata              | ``AutoModerationRuleTriggerMetadata``                     |
            +-------------------------------+-----------------------------------------------------------+
            | trigger_type                  | ``AutoModerationRuleTriggerType``                         |
            +-------------------------------+-----------------------------------------------------------+
        """
        old_attributes = {}
        
        # id
        # Already set & cannot change
        
        # actions
        action_array = data['actions']
        if action_array:
            actions = tuple(
                AutoModerationAction.from_data(auto_moderation_action_data)
                for auto_moderation_action_data in action_array
            )
        else:
            actions = None
        
        if self.actions != actions:
            old_attributes['actions'] = self.actions
            self.actions = actions
        
        # creator_id
        # Internal field & cannot change
        
        # enabled
        enabled = data['enabled']
        if self.enabled != enabled:
            old_attributes['enabled'] = self.enabled
            self.enabled = enabled
        
        # event_type
        event_type = AutoModerationEventType.get(data['event_type'])
        if self.event_type != event_type:
            old_attributes['event_type'] = self.event_type
            self.event_type = event_type
        
        # excluded_channel_ids
        excluded_channel_id_array = data['exempt_channels']
        if excluded_channel_id_array:
            excluded_channel_ids = tuple(sorted(int(channel_id) for channel_id in excluded_channel_id_array))
        else:
            excluded_channel_ids = None
        
        if self.excluded_channel_ids != excluded_channel_ids:
            old_attributes['excluded_channel_ids'] = self.excluded_channel_ids
            self.excluded_channel_ids = excluded_channel_ids
        
        # excluded_role_ids
        excluded_role_id_array = data['exempt_roles']
        if excluded_role_id_array:
            excluded_role_ids = tuple(sorted(int(role_id) for role_id in excluded_role_id_array))
        else:
            excluded_role_ids = None
        
        if self.excluded_role_ids != excluded_role_ids:
            old_attributes['excluded_role_ids'] = self.excluded_role_ids
            self.excluded_role_ids = excluded_role_ids
        
        # guild_id
        # Internal field & cannot change
        
        # name
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # trigger_metadata & trigger_type
        trigger_type = AutoModerationRuleTriggerType.get(data['trigger_type'])
        trigger_metadata_type = trigger_type.metadata_type
        if (trigger_metadata_type is None):
            trigger_metadata = None
        else:
            trigger_metadata = trigger_metadata_type.from_data(data['trigger_metadata'])
            
        if self.trigger_metadata != trigger_metadata:
            old_attributes['trigger_metadata'] = self.trigger_metadata
            self.trigger_metadata = trigger_metadata
        
        if self.trigger_type is not trigger_type:
            old_attributes['trigger_type'] = self.trigger_type
            self.trigger_type = trigger_type
        
        return old_attributes
    
    
    def __eq__(self, other):
        """Returns whether the two auto moderation rules are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id == other_id
        
        # actions
        if self.actions != other.actions:
            return False
        
        # creator_id
        # Internal field
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # event_type
        if self.event_type is not other.event_type:
            return False
        
        # excluded_channel_ids
        if self.excluded_channel_ids != other.excluded_channel_ids:
            return False
        
        # excluded_role_ids
        if self.excluded_role_ids != other.excluded_role_ids:
            return False
        
        # guild_id
        # Internal field
        
        # name
        if self.name != other.name:
            return False
        
        # trigger_metadata
        if self.trigger_metadata != other.trigger_metadata:
            return False
        
        # trigger_type
        if self.trigger_type is not other.trigger_type:
            return False
        
        return True
    
    
    def __ne__(self, other):
        """Returns whether the two auto moderation rules are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # id
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return self_id != other_id
        
        # actions
        if self.actions != other.actions:
            return True
        
        # creator_id
        # Internal field
        
        # enabled
        if self.enabled != other.enabled:
            return True
        
        # event_type
        if self.event_type is not other.event_type:
            return True
        
        # excluded_channel_ids
        if self.excluded_channel_ids != other.excluded_channel_ids:
            return True
        
        # excluded_role_ids
        if self.excluded_role_ids != other.excluded_role_ids:
            return True
        
        # guild_id
        # Internal field
        
        # name
        if self.name != other.name:
            return True
        
        # trigger_metadata
        if self.trigger_metadata != other.trigger_metadata:
            return True
        
        # trigger_type
        if self.trigger_type is not other.trigger_type:
            return True
        
        return False
    
    
    def __hash__(self):
        """Returns the auto moderation rule's hash value."""
        
        # id
        self_id = self.id
        if self_id:
            return self_id
        
        hash_value = 0
        
        # actions
        actions = self.actions
        if (actions is not None):
            hash_value ^= len(actions)
            
            for action in actions:
                hash_value ^= hash(action)
        
        # creator_id
        # Internal field
        
        # enabled
        hash_value ^= self.enabled << 4
        
        # event_type
        hash_value ^= self.event_type.value << 8
        
        # excluded_channel_ids
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            hash_value ^= len(excluded_channel_ids) << 12
            
            for excluded_channel_id in excluded_channel_ids:
                hash_value ^= excluded_channel_id
        
        # excluded_role_ids
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            hash_value ^= len(excluded_role_ids) << 16
            
            for excluded_role_id in excluded_role_ids:
                hash_value ^= excluded_role_id
        
        # guild_id
        # Internal field
        
        # name
        hash_value ^= hash(self.name)
                
        # trigger_metadata1
        trigger_metadata = self.trigger_metadata
        if (trigger_metadata is not None):
            hash_value ^= hash(trigger_metadata)
        
        # trigger_type
        hash_value ^= self.trigger_type.value << 20
        
        return hash_value
    
    
    def __repr__(self):
        """Returns the auto moderation rule's representation"""
        repr_parts = ['<', self.__class__.__name__]
        
        # system fields: `.id`, `.guild_id`, `.creator_id`
        
        # id
        id_ = self.id
        if id_ == 0:
            repr_parts.append(' partial')
        
        else:
            repr_parts.append(' id=')
            repr_parts.append(repr(id_))
            
            repr_parts.append(', guild_id=')
            repr_parts.append(repr(self.guild_id))
            
            repr_parts.append(', creator_id=')
            repr_parts.append(repr(self.creator_id))
        
        # Information `.name`, `.trigger_type`
        
        # name
        repr_parts.append(', name=')
        repr_parts.append(repr(self.name))
        
        # trigger_type
        trigger_type = self.trigger_type
        repr_parts.append(', trigger_type=')
        repr_parts.append(repr(trigger_type.name))
        repr_parts.append('~')
        repr_parts.append(repr(trigger_type.value))
        
        # Additional: `.actions`, `.enabled`, `event_type`, `.excluded_channel_ids`, `.excluded_role_ids`, 
        # `.trigger_metadata`
        
        trigger_metadata = self.trigger_metadata
        if (trigger_metadata is not None):
            repr_parts.append(', trigger_metadata=')
            repr_parts.append(repr(trigger_metadata))
        
        repr_parts.append(', actions=[')
        actions = self.actions
        if (actions is not None):
            limit = len(actions)
            index = 0
            
            while True:
                action = actions[index]
                repr_parts.append(repr(action))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append(']')
        
        # enabled
        enabled = self.enabled
        if not enabled:
            repr_parts.append(', enabled=')
            repr_parts.append(repr(enabled))
        
        event_type = self.event_type
        if (
            (event_type is not AutoModerationEventType.none) and
            (event_type is not AutoModerationEventType.message_send)
        ):
            repr_parts.append(', event_type=')
            event_type.append(repr(event_type.name))
            event_type.append('~')
            event_type.append(repr(event_type.value))
        
        # excluded_channel_ids
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            repr_parts.append(', excluded_channel_ids=[')
            
            limit = len(excluded_channel_ids)
            index = 0
            
            while True:
                excluded_channel_id = excluded_channel_ids[index]
                repr_parts.append(repr(excluded_channel_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # excluded_role_ids
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            repr_parts.append(', excluded_role_ids=[')
            
            limit = len(excluded_role_ids)
            index = 0
            
            while True:
                excluded_role_id = excluded_role_ids[index]
                repr_parts.append(repr(excluded_role_id))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the auto moderation rule returning a new partial one.
        
        Returns
        -------
        new : ``AutoModerationRule``
        """
        new = object.__new__(type(self))
        
        # id
        # Internal field
        new.id = 0
        
        # actions
        actions = self.actions
        if (actions is not None):
            actions = tuple(action.copy() for action in actions)
        new.actions = actions
        
        # creator_id
        # Internal field
        new.creator_id = 0
        
        # enabled
        new.enabled = self.enabled
        
        # event_type
        new.event_type = self.event_type
        
        # excluded_channel_ids
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            excluded_channel_ids = tuple(excluded_channel_id for excluded_channel_id in excluded_channel_ids)
        new.excluded_channel_ids = excluded_channel_ids
        
        # excluded_role_ids
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            excluded_role_ids = tuple(excluded_role_id for excluded_role_id in excluded_role_ids)
        new.excluded_role_ids = excluded_role_ids
        
        # guild_id
        # Internal field
        new.guild_id = 0
        
        # name
        new.name = self.name
        
        # trigger_metadata
        trigger_metadata = self.trigger_metadata
        if (trigger_metadata is not None):
            trigger_metadata = trigger_metadata.copy()
        new.trigger_metadata = trigger_metadata
        
        # trigger_type
        new.trigger_type = self.trigger_type
        
        return new
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the auto moderation rule with the given attributes replaced.
        
        Parameters
        ----------
        actions : `None`, ``AutoModerationAction``, `iterable` of ``AutoModerationAction``, Optional (Keyword only)
            Actions which will execute when the rule is triggered.
        
        enabled : `bool` = `True`, Optional (Keyword only)
            Whether the rule is enabled.
        
        event_type : `None`, `int`, ``AutoModerationEventType`` = `AutoModerationEventType.message_send` \
                , Optional (Keyword only)
            For which events is the rule applied.
        
        excluded_channels : `None`, `int`, ``Channel``, `iterable` of (`int`, ``Channel``) = `None` \
                , Optional (Keyword only)
            Excluded channels from the rule.
        
        excluded_keywords : `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            Excluded keywords from preset filter.
            
            > Only meaningful with the `keyword_presets` parameter.
        
        excluded_roles : `None`, `int`, ``Role``, `iterable` of (`int`, ``Role``) = `None`, Optional (Keyword only)
            Excluded roles from the rule.
        
        keyword_presets : `None`, `int`, ``AutoModerationKeywordPresetType``, \
                `iterable` of (`int`, ``AutoModerationKeywordPresetType``), Optional (Keyword only)
            Keyword preset defined by Discord which will be searched for in content.
            
            > Mutually exclusive with the `keywords` and `mention_limit` parameters.
        
        keywords : `None`, `str`, `iterable` of `str`, Optional (Keyword only)
            Substrings which will be searched for in content.
            
            > Mutually exclusive with the `keyword_presets` and `mention_limit` parameters.
        
        mention_limit : `None`, `int`, Optional (Keyword only)
            The amount of mentions in a message after the rule is triggered.
            
            > Mutually exclusive with the `keyword_presets` and `keywords` parameters.
        
        name : `str`, Optional (Keyword only)
            The rule's name.
        
        trigger_type : ``AutoModerationRuleTriggerType``, `int`, Optional (Keyword only)
            Auto moderation trigger type.
            
            By passing `keyword_presets` parameter you can define the trigger type as 
            `AutoModerationRuleTriggerType.keyword_preset`, or by passing the `keywords` you can define it as 
            `AutoModerationRuleTriggerType.keyword`.
        
        Returns
        -------
        new : ``AutoModerationRule``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - If a parameter is unexpected.
        ValueError
            - If a parameter's value is incorrect.
        """
        # id
        # Internal field
        
        # actions
        try:
            actions = keyword_parameters.pop('actions')
        except KeyError:
            actions = self.actions
            if (actions is not None):
                actions = tuple(action.copy() for action in actions)
        else:
            actions = _validate_actions(actions)
        
        # creator_id
        # internal field
        
        # enabled
        try:
            enabled = keyword_parameters.pop('enabled')
        except KeyError:
            enabled = self.enabled
        else:
            enabled = _validate_enabled(enabled)
        
        # enabled
        try:
            event_type = keyword_parameters.pop('event_type')
        except KeyError:
            event_type = self.event_type
        else:
            event_type = _validate_event_type(event_type)
        
        # excluded_channel_ids
        try:
            excluded_channels = keyword_parameters.pop('excluded_channels')
        except KeyError:
            
            excluded_channel_ids = self.excluded_channel_ids
            if (excluded_channel_ids is not None):
                excluded_channel_ids = tuple(excluded_channel_id for excluded_channel_id in excluded_channel_ids)
        
        else:
            excluded_channel_ids = _validate_excluded_channel_ids(excluded_channels)
        
        # excluded_role_ids
        try:
            excluded_roles = keyword_parameters.pop('excluded_roles')
        except KeyError:
            
            excluded_role_ids = self.excluded_role_ids
            if (excluded_role_ids is not None):
                excluded_role_ids = tuple(excluded_role_id for excluded_role_id in excluded_role_ids)
        
        else:
            excluded_role_ids = _validate_excluded_role_ids(excluded_roles)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            name = self.name
        
        else:
            name = _validate_name(name)
        
        # trigger_metadata & trigger_type
        trigger_type = keyword_parameters.pop('trigger_type', ...)
        excluded_keywords = keyword_parameters.pop('excluded_keywords', ...)
        keyword_presets = keyword_parameters.pop('keyword_presets', ...)
        keywords = keyword_parameters.pop('keywords', ...)
        mention_limit = keyword_parameters.pop('mention_limit', ...)
        
        if (trigger_type is ...) and (keyword_presets is ...) and (keywords is ...) and (mention_limit is ...):
            trigger_metadata = self.trigger_metadata
            if (trigger_metadata is not None):
                trigger_metadata = trigger_metadata.copy()
            
            trigger_type = self.trigger_type
            
        else:
            trigger_metadata, trigger_type = _validate_trigger_type_with_metadata_options(
                trigger_type, excluded_keywords, keyword_presets, keywords, mention_limit
            )
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        new = object.__new__(type(self))
        
        new.id = 0
        new.actions = actions
        new.creator_id = 0
        new.enabled = enabled
        new.event_type = event_type
        new.excluded_channel_ids = excluded_channel_ids
        new.excluded_role_ids = excluded_role_ids
        new.guild_id = 0
        new.name = name
        new.trigger_metadata = trigger_metadata
        new.trigger_type = trigger_type
        
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the auto moderation rule is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (not self.id)
    
    
    def iter_actions(self):
        """
        Iterates over the actions of the auto moderation rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        action : ``AutoModerationAction``
        """
        actions = self.actions
        if (actions is not None):
            yield from actions
    
    
    @property
    def creator(self):
        """
        Returns who created the auto moderation rule.
        
        Returns
        -------
        creator : ``ClientUserBase``
        """
        creator_id = self.creator_id
        if creator_id:
            creator = create_partial_user_from_id(creator_id)
        
        else:
            creator = ZEROUSER
        
        return creator
    
    
    def iter_excluded_channel_ids(self):
        """
        Iterates over the excluded channel identifiers from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_channel_id : `int`
        """
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            yield from excluded_channel_ids
    
    
    @property
    def excluded_channels(self):
        """
        Returns the excluded channels from the rule.
        
        Returns
        -------
        excluded_channels : `None`, `tuple` of ``Channel``
        """
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is None):
            excluded_channels = None
        
        else:
            guild_id = self.guild_id
            excluded_channels = tuple(
                create_partial_channel_from_id(excluded_channel_id, -1, guild_id)
                for excluded_channel_id in excluded_channel_ids
            )
        
        return excluded_channels
    
    
    def iter_excluded_channels(self):
        """
        Iterates over the excluded channels from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_channel : ``Channel``
        """
        excluded_channel_ids = self.excluded_channel_ids
        if (excluded_channel_ids is not None):
            guild_id = self.guild_id
            
            for excluded_channel_id in excluded_channel_ids:
                yield create_partial_channel_from_id(excluded_channel_id, -1, guild_id)
    
    
    def iter_excluded_role_ids(self):
        """
        Iterates over the excluded role identifiers from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_role_id : `int`
        """
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            yield from excluded_role_ids
    
    
    @property
    def excluded_roles(self):
        """
        Returns the excluded roles from the rule.
        
        Returns
        -------
        excluded_roles : `None`, `tuple` of ``Role``
        """
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is None):
            excluded_roles = None
        
        else:
            excluded_roles = tuple(
                create_partial_role_from_id(excluded_role_id) for excluded_role_id in excluded_role_ids
            )
        
        return excluded_roles
    
    
    def iter_excluded_roles(self):
        """
        Iterates over the excluded roles from the rule.
        
        This method is an iterable generator.
        
        Yields
        ------
        excluded_role : ``Role``
        """
        excluded_role_ids = self.excluded_role_ids
        if (excluded_role_ids is not None):
            for excluded_role_id in excluded_role_ids:
                yield create_partial_role_from_id(excluded_role_id)



__all__ = ()

from ...discord import Event, EventHandlerPlugin

from . import track_end_reasons as TRACK_END_REASONS


async def default_track_exception_event_handler(client, event):
    """
    Handles track exception by starting to play the next one.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    event : ``TrackExceptionEvent``
        The exception event received.
    """
    await event.player.remove(0)


async def default_track_stuck_event_handler(client, event):
    """
    Handles track stuck by starting to play the next one.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective event.
    event : ``TrackStuckEvent``
        The track stuck event.
    """
    await event.player.skip(0)


async def default_track_end_event_handler(client, event):
    """
    Handles track end by starting to play the next one.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective event.
    event : ``TrackEndEvent``
        The track end event.
    """
    if event.reason == TRACK_END_REASONS.finished:
        await event.player.skip(0)


async def default_player_websocket_closed_event_handler(client, event):
    """
    Handles websocket close event.
    
    This method is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    event : ``PlayerWebsocketClosedEvent``
        The received websocket closed event.
    """
    player = event.player
    if event.should_reconnect():
        if player.is_playing():
            current = player.get_current()
            if (current is not None):
                current = current.copy()
                current.start_time = player.position
                await player._play(current)
    
    else:
        player.disconnect()


class SolarLinkEventManager(EventHandlerPlugin):
    """
    Plugin to extend `client.events` auto registered when setting up the extension to a client.
    
    Event Attributes
    ----------------
    track_end(client: ``Client``, event: ``TrackEndEvent``)
        Called when a track ended.
    
    track_exception(client: ``Client``, event: ``TrackExceptionEvent``)
        Called when a track finished with exception.
    
    track_start(client: ``Client``, event: ``TrackStartEvent``)
        Called when a track started.
    
    track_stuck(client: ``Client``, event: ``TrackStuckEvent``)
        Called when the currently playing track is stuck.
    
    player_websocket_closed(client: ``Client``, event: ``PlayerWebsocketClosedEvent``)
        Called when a player's websocket is disconnected from a guild.
    """
    track_end = Event(2, default_handler=default_track_end_event_handler)
    track_exception = Event(2, default_handler=default_track_exception_event_handler)
    track_start = Event(2)
    track_stuck = Event(2, default_handler=default_track_stuck_event_handler)
    player_websocket_closed = Event(2, default_handler=default_player_websocket_closed_event_handler)
    
    def __repr__(self):
        """Returns the plugin's representation."""
        return f'<{self.__class__.__name__}>'



__all__ = ()

from scarletio import Compound
from ....env import API_VERSION

from ...bases import maybe_snowflake
from ...http import DiscordHTTPClient
from ...integration import Integration
from ..request_helpers import get_guild_id


class ClientCompoundIntegrationEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    #TODO: decide if we should store integrations at Guild objects
    if API_VERSION > 7:
        async def integration_get_all(self, guild):
            """
            Requests the integrations of the given guild.
            
            This method is a coroutine.
            
            Parameters
            ----------
            guild : ``Guild``, `int`
                The guild, what's integrations will be requested.
            
            Returns
            -------
            integrations : `list` of ``Integration``
            
            Raises
            ------
            TypeError
                If `guild` was not given neither as ``Guild`` nor `int`.
            ConnectionError
                No internet connection.
            DiscordException
                If any exception was received from the Discord API.
            """
            guild_id = get_guild_id(guild)
            
            integration_datas = await self.http.integration_get_all(guild_id, None)
            return [Integration(integration_data) for integration_data in integration_datas]
    
    else:
        async def integration_get_all(self, guild):
            """
            Requests the integrations of the given guild.
            
            This method is a coroutine.
            
            Parameters
            ----------
            guild : ``Guild``, `int`
                The guild, what's integrations will be requested.
            
            Returns
            -------
            integrations : `list` of ``Integration``
            
            Raises
            ------
            TypeError
                If `guild` was not given neither as ``Guild`` nor `int`.
            ConnectionError
                No internet connection.
            DiscordException
                If any exception was received from the Discord API.
            """
            guild_id = get_guild_id(guild)
            
            integration_datas = await self.http.integration_get_all(guild_id, {'include_applications': True})
            return [Integration(integration_data) for integration_data in integration_datas]
    
    
    async def integration_create(self, guild, integration_id, type_):
        """
        Creates an integration at the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to what the integration will be attached to.
        integration_id : ``int``
            The integration's id.
        type_ : `str`
            The integration's type (`'twitch'`, `'youtube'`, etc.).
        
        Returns
        -------
        integration : ``Integration``
            The created integrated.
        
        Raises
        ------
        TypeError
            - If `guild` was not given neither as ``Guild`` nor `int`.
            - If `integration_id` was not given as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `type_` is not given as `str`.
        """
        guild_id = get_guild_id(guild)
        
        integration_id_value = maybe_snowflake(integration_id)
        if integration_id_value is None:
            raise TypeError(
                f'`integration_id` can be `int`, got {integration_id.__class__.__name__}; {integration_id!r}.'
            )
        
        if __debug__:
            if not isinstance(type_, str):
                raise AssertionError(
                    f'`type_` can be `str`, got {type_.__class__.__name__}; {type_!r}.'
                )
        
        data = {
            'id'   : integration_id_value,
            'type' : type_,
        }
        
        data = await self.http.integration_create(guild_id, data)
        return Integration(data)
    
    
    async def integration_edit(
        self, integration, *, expire_behavior=..., expire_grace_period=..., enable_emojis=...
    ):
        """
        Edits the given integration.
        
        This method is a coroutine.
        
        Parameters
        ----------
        integration : ``Integration``
            The integration to edit.
        expire_behavior : `int`, Optional (Keyword only)
            Can be `0` for kick or `1` for role  remove.
        expire_grace_period : `int`, Optional (Keyword only)
            The time in days, after the subscription will be ignored. Can be any of `(1, 3, 7, 14, 30)`.
        enable_emojis : `bool`, Optional (Keyword only)
            Whether the users can use the integration's emojis in Discord.
        
        Raises
        ------
        TypeError
            - If `expire_behavior` was not passed as `int`.
            - If `expire_grace_period` was not passed as `int`.
            - If `enable_emojis` was not passed as `bool`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `integration` was not given as ``Integration``.
            - If `expire_behavior` was not given neither as `None` nor as `int`.
            - If `expire_grace_period` was not given neither as `None` nor as `int`.
            - If `expire_behavior` is not any of: `(0, 1)`.
            - If `expire_grace_period` is not any of `(1, 3, 7, 14, 30)`.
            - If `enable_emojis` is neither `None`, `bool`.
        """
        if __debug__:
            if not isinstance(integration, Integration):
                raise AssertionError(
                    f'`integration` can be `{Integration.__name__}`, got '
                    f'{integration.__class__.__name__}; {integration!r}.'
                )
        
        detail = integration.detail
        if detail is None:
            return
        
        role = detail.role
        if role is None:
            return
        
        data = {}
        
        if expire_behavior is not ...:
            if __debug__:
                if not isinstance(expire_behavior, int):
                    raise AssertionError(
                        f'`expire_behavior` can be `None`, `int`, got '
                        f'{expire_behavior.__class__.__name__}; {expire_behavior!r}.'
                    )
                
                if expire_behavior not in (0, 1):
                    raise AssertionError(
                        f'`expire_behavior` should be 0 for kick, 1 for remove role, got {expire_behavior!r}.'
                    )
            
            data['expire_behavior'] = expire_behavior
        
        if expire_grace_period is not ...:
            if __debug__:
                if not isinstance(expire_grace_period, int):
                    raise AssertionError(
                        f'`expire_grace_period` can be `None`, `int`, got '
                        f'{expire_grace_period.__class__.__name__}.'
                    )
                
                if expire_grace_period not in (1, 3, 7, 14, 30):
                    raise AssertionError(
                        f'`expire_grace_period` can be one of `(1, 3, 7, 14, 30)`, got {expire_grace_period!r}.'
                    )
                
            data['expire_grace_period'] = expire_grace_period
   
        
        if (enable_emojis is not ...):
            if __debug__:
                if not isinstance(enable_emojis, bool):
                    raise AssertionError(
                        f'`enable_emojis` can be `None`, `bool`, got '
                        f'{enable_emojis.__class__.__name__}; {enable_emojis!r}.'
                    )
            
            data['enable_emoticons'] = enable_emojis
        
        await self.http.integration_edit(role.guild_id, integration.id, data)
    
    
    async def integration_delete(self, integration):
        """
        Deletes the given integration.
        
        This method is a coroutine.
        
        Parameters
        ----------
        integration : ``Integration``
            The integration what will be deleted.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `integration` was not given as ``Integration``.
        """
        if __debug__:
            if not isinstance(integration, Integration):
                raise AssertionError(
                    f'`integration` can be `{Integration.__name__}`, got {integration.__class__.__name__}; '
                    f'{integration!r}.'
                )
        
        detail = integration.detail
        if detail is None:
            return
        
        role = detail.role
        if role is None:
            return
        
        await self.http.integration_delete(role.guild_id, integration.id)
    
    
    async def integration_sync(self, integration):
        """
        Sync the given integration's state.
        
        This method is a coroutine.
        
        Parameters
        ----------
        integration : ``Integration``
            The integration to sync.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `integration` was not given as ``Integration``.
        """
        if __debug__:
            if not isinstance(integration, Integration):
                raise AssertionError(
                    f'`integration` can be `{Integration.__name__}`, got {integration.__class__.__name__}; '
                    f'{integration!r}.'
                )
        
        detail = integration.detail
        if detail is None:
            return
        
        role = detail.role
        if role is None:
            return
        
        await self.http.integration_sync(role.guild_id, integration.id)



__all__ = ('PluginModuleProxyType',)

import warnings
from types import ModuleType

from scarletio import get_last_module_frame, include

from ..constants import PLUGINS

PluginModuleSpecType = include('PluginModuleSpecType')


class PluginModuleProxyType(ModuleType):
    """
    Represents an plugin module's proxy.
    """
    def __init__(self, spec):
        """
        Creates a new plugin module proxy.
        
        Parameters
        ----------
        spec : ``PluginModuleSpecType``
            Module specification.
        """
        ModuleType.__init__(self, spec.name)
        
        self.__spec__ = spec
        self.__file__ = spec.origin
    
    
    def __setattr__(self, attribute_name, attribute_value):
        ModuleType.__setattr__(self, attribute_name, attribute_value)
        
        if attribute_name in {
            '__spec__',
            '__loader__',
            '__package__',
            '__spec__',
            '__path__',
            '__file__',
            '__cached__',
        }:
            pass
        
        elif isinstance(attribute_value, PluginModuleProxyType):
            spec = self.__spec__
            
            module = spec.get_module()
            if (module is not None):
                setattr(module, attribute_name, attribute_value)
        
        else:
            warnings.warn(
                f'Unallowed attribute assignment: `{attribute_name} = {attribute_value!r}` of type '
                f'`{type(attribute_value).__name__}` to `{self.__spec__.name}`'
            )
    
    
    def __getattr__(self, attribute_name):
        spec = self.__spec__
        module = spec.get_module()
        
        if module is None:
            raise RuntimeError(
                f'Plugin, `{spec.name}` is not yet initialized!'
            )
        
        try:
            attribute_value = getattr(module, attribute_name)
        except AttributeError:
            raise AttributeError(
                f'{self!r} has no attribute `{attribute_name}`.'
            ) from None
        
        frame = get_last_module_frame()
        if (frame is None):
            spec = None
        else:
            spec = frame.f_globals.get('__spec__', None)
        
        if spec is None:
            plugin = None
        else:
            plugin = PLUGINS.get(spec.name)
        
        current_plugin = PLUGINS.get(spec.name, None)
        
        if (plugin is not None) and (current_plugin is not None) and (plugin is not current_plugin):
            plugin.add_child_plugin(current_plugin)
            current_plugin.add_parent_plugin(plugin)
        
        return attribute_value
    
    
    def __repr__(self):
        spec = getattr(self, '__spec__', None)
        
        repr_parts = ['<', type(self).__name__]
        
        if spec is None:
            repr_parts.append(' unknown location')
        
        else:
            repr_parts.append(' from ')
            repr_parts.append(spec.name)
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __dir__(self):
        directory = {*ModuleType.__dir__(self)}
        
        module = self.__spec__.get_module()
        if (module is not None):
            directory.update(dir(module))
        
        return sorted(directory)
